'''module to set up particle phase part of box model, calls on init_water_partit to initiate water partitioning with seed particles and wall'''
# using the user-defined or default values, the initial number size distribution is determined here

import numpy as np
import size_distr # custom library - see source code
from init_water_partit import init_water_partit
import scipy.constants as si

def pp_intro(y, num_comp, Pybel_objects, TEMP, H2Oi,
		mfp, accom_coeff, y_mw, surfT, 
		RH, siz_str, num_asb, lowersize, uppersize, pmode, pconc, 
		pconct, nuc_comp, testf, std, mean_rad, therm_sp,
		y_dens, Psat, core_diss, kgwt, space_mode, seedVr, 
		spec_namelist, act_coeff, wall_on, partit_cutoff, Press,
		seedi):
	
	# inputs -----------------------------------
	# TEMP - temperature (K) in chamber at start of experiment
	# y_mw - molecular weight (g/mol) of components (num_comp, 1)
	# siz_str - the size structure
	# num_asb - number of size bins (excluding wall)
	# lowersize - lowest size bin radius bound (um)
	# uppersize - largest size bin radius bound (um)
	# pmode - whether particle number concentrations given as modes or explicitly
	# pconc - starting particle concentration (# particle/cc (air)) - if scalar then
	# gets split between size bins in Size_distributions call, or if an array, elements 
	# are allocated to corresponding size bins
	# pconct - time(s) through experiment that particle number concentrations 
	#	correspond to (s)
	# nuc_comp - name of the nucleating component
	# testf - test flag to say whether in normal mode (0) or test mode for front.py (1)
	#       or test mode for pp_intro.py
	# std - geometric standard deviation of the particle number concentration 
	# 		(dimensionless)
	# mean_rad - either the mean radius (um) of particles in lognormal number-size 
	#			distribution (in which case pconc should be scalar), or mean radius of
	#			particles where just one size bin present (in which case pconc is also
	#			scalar)
	# y_dens - liquid density of components (kg/m3) (num_comp, 1)
	# Psat - saturation vapour pressure of components (molecules/cc (air))
	# core_diss - core dissociation constant
	# kgwt - mass transfer coefficient for vapour-wall partitioning (/s)
	# space_mode - string specifying whether to space size bins logarithmically or 
	# linearly
	# seedVr - volume ratio of component(s) comprising seed particles
	# spec_namelist - names of components noted in chemical scheme file
	# act_coeff - activity coefficient of components
	# wall_on - whether or not to consider wall
	# partit_cutoff - product of vapour pressure and activity coefficient
	#		at which gas-particle partitioning assumed zero (Pa)
	# Press - pressure inside chamber
	# seedi - index of seed components
	# ------------------------------------------
	
	if testf==1: # in test mode
		return(0,0,0,0,0,0,0,0,0,0,0,0) # return dummies
	
	# isolate the starting number size distribution information
	i = (pconct[0, :] == 0) # index of initial information
	
	if (sum(i) == 0): # if no initial information provide fillers
		pconcn = np.zeros((1))
		stdn = [1.e20]
		mean_radn = [-1.e6]
	else: # obtain initial information
		pconcn = pconc[:, i]

		try: # if mean_rad already an array
			if (std[:, i].ndim > 1):
				stdn = std[:, i]
			if (std[:, i].ndim == 1):
				stdn = std[:, i]
			if (std[:, i].ndim == 0):
				stdn = [std[:, i]]
		except:
			stdn = [std[:, i]]
		
		try: # if mean_rad already an array
			if (mean_rad[:, i].ndim > 1):
				mean_radn = mean_rad[:, i]
			if (mean_rad[:, i].ndim == 1):
				mean_radn = mean_rad[:, i]
			if (mean_rad[:, i].ndim == 0):
				mean_radn = [mean_rad[:, i]]
		except:
			mean_radn = [mean_rad[:, i]]
	
	# if mean radius not stated explicitly calculate from size ranges (um)
	if (num_asb > 0):
		if (any(mrn == -1.e6 for mrn in mean_radn)):
			if (lowersize > 0.):
				mean_radn[mean_radn == -1.e6] = [10**((np.log10(lowersize)+np.log10(uppersize))/2.0)]
			if (lowersize == 0.):
				mean_radn[mean_radn == -1.e6] = [10**((np.log10(uppersize))/2.0)]
	
	# index of nucleating component
	if len(nuc_comp)>0:
		nuc_compi = spec_namelist.index(nuc_comp[0])
		nuc_comp = np.empty(1, dtype=int)
		nuc_comp[0] = nuc_compi

	R_gas = si.R # ideal gas constant (kg.m2.s-2.K-1.mol-1)
	NA = si.Avogadro # Avogadro's number (molecules/mol)
	
	
	if num_asb == 0: # create dummy variables if no size bins
		N_perbin = np.zeros((1,1))
		x = np.zeros((1,1))
		Varr = np.zeros((1,1))
		Vbou = np.zeros((1,1))
		rad0 = np.zeros((1,1))
		Vol0 = np.zeros((1,1))
		rbou = np.zeros((1,1))
		upper_bin_rad_amp = 1.0e6
	
	# create a number concentration for a lognormal distribution (particles/cc (air))
	# this is where gas partitioning to wall set up
	if (testf == 2):
		print('calling Size_distributions.lognormal')
		
	# if multiple size bins, this call will assume a lognormal distribution if initial 
	# particle concentration is described by mode, or will assign particles to size bins if
	# initial particle concentration per size bin provided
	if (num_asb > 1):
			
		# set scale and standard deviation input for lognormal probability distribution 
		# function, following guidance here: 
		# http://all-geo.org/volcan01010/2013/09/how-to-use-lognormal-distributions-in-python/
		# if mean_radn and stdn are not already arrays then transform to a list 
		if type(mean_radn) != np.ndarray:
			scale = [np.exp(np.log(mean_radn))]
		else:
			scale = np.exp(np.log(mean_radn))
		
		if type(stdn) != np.ndarray:
			stdn = [np.log(stdn)]
		else:
			stdn = np.log(stdn)
			
		loc = 0. # no shift
		
		[N_perbin, x, rbou, Vbou, Varr, upper_bin_rad_amp] = size_distr.lognormal(num_asb, 
			pmode, pconcn, stdn, lowersize, uppersize, loc, scale, space_mode)
		

		if (testf == 2):
			print('finished with Size_distributions.lognormal')
			
	if (num_asb == 1):

		N_perbin = np.array((sum(pconcn))).reshape(-1, 1) # (# particles/cc (air))
		x = np.zeros(1) # radii at size bin centre
		# mean radius of this one size bin (um)
		if (len(mean_radn[0]) == 1): # if a scalar
			meansize = mean_radn[0][0]
		else : # if an array
			meansize = sum(mean_radn)/len(mean_radn)
			
		x[0] = meansize

		# extend uppersize to reduce chance of particles growing beyond this
		upper_bin_rad_amp = 1.0e6
		uppersize = uppersize*upper_bin_rad_amp
		# volume bounds of size bin (um3)
		Vbou = np.array(((lowersize**3.0)*(4.0/3.0)*np.pi, 
						(uppersize**3.0)*(4.0/3.0)*np.pi))
		# volume of single particle (um3)
		Varr = np.zeros((1, 1))
		Varr[0] = (4./3.)*np.pi*(x[0]**3.0)
		# radius bounds of size bin (um)
		rbou = ((Vbou*3.0)/(4.0*np.pi))**(1.0/3.0)

	# set first volume and radius bound to zero, thereby allowing shrinkage to zero in the 
	# smallest bin
	# remember initial first radius bound for saving
	rbou00 = rbou[0]
	Vbou[0] = 0.
	rbou[0] = 0. # this reversed in saving.py back to rbou00
	
	if (num_asb > 0):
		# remember the radii (um) and volumes (um3) at size bin centre before water 
		# partitioning
		rad0 = np.zeros((len(x)))
		rad0[:] = x[:]
		Vol0 = np.zeros((len(Varr)))
		Vol0[:] = Varr[:]
	
	if (wall_on > 0):
		num_asb += 1 # add one to size bin number to account for wall

	# append particle and wall concentrations of components to y (molecules/cc (air))
	y = np.append(y, np.zeros((num_asb*num_comp)))
	
	# molar volume (multiply y_dens by 1e-3 to convert from kg/m3 to g/cc and give
	# MV in units cc/mol)
	MV = (y_mw/(y_dens*1.0e-3)).reshape(num_comp, 1)
	if (sum(pconcn) > 0.0): # account for concentration of components comprising seed
		for ci in range(len(seedi)): # loop through indices of seed components
			# concentration in all size bins (molecules/cc (air)):
			y[num_comp+seedi[ci]:(num_comp*(num_asb)+seedi[ci]):num_comp] = (NA/MV[seedi[ci]])*(Varr*1.e-12*(seedVr[ci]/sum(seedVr)))*N_perbin[:, 0]
	
	# print mass concentration of particles (scale y_dens by 1e-3 to convert from kg/m3
	# to g/cm3)
	if (num_asb > 0): # with particles
		mass_conc = 0. # start cumulation
		for i in range(num_asb-1): # as size bin now account for wall too
			mass_conc += sum((y_dens[:, 0]*1.0e-3)*((y[num_comp*(i+1):num_comp*(i+2)]/si.N_A)*MV[:,0]))
			mass_conc -= (y_dens[int(H2Oi), 0]*1.0e-3)*((y[num_comp*(i+1)+int(H2Oi)]/si.N_A)*MV[int(H2Oi), 0])
		mass_conc = mass_conc*1.e12 # convert from g/cc (air) to ug/m3 (air)
		
		print(str('Total dry (no water) mass concentration of particles at start of simulation is ' + str(mass_conc) + ' ug/m3 (air)'))
	else: # no particles
		print('No particle size bins detected, simulation will not include particles')

	# start counter on number concentration of newly nucleated particles (#/cc (air))
	np_sum = 0.
	
	return(y, N_perbin, x, Varr, Vbou, rad0, Vol0, rbou, MV, num_asb, nuc_comp, rbou00, 
			upper_bin_rad_amp, np_sum)



'''interrogate the xml file'''
# opens and extracts the component names and associated SMILE strings
# from the xml file

import xmltodict # for opening and converting xml files to python dictionaries

# define function
def xml_interr(xml_name):

	# inputs: --------------------------------------------------------
	# xml_name - name of xml file
	# ----------------------------------------------------------------

	with open(xml_name) as fd:
		doc = xmltodict.parse(fd.read())

	a = doc['mechanism']['species_defs']['species']
	
	# prepare arrays to fill	
	comp_numb = list(('0',) * len(a))
	comp_name = list(('0',) * len(a))
	comp_smil = list(('0',) * len(a))
	
	for i in range(len(a)):
		comp_numb[i] = a[i]['@species_number']
		comp_name[i] = a[i]['@species_name']
		if "smiles" in a[i]:
			comp_smil[i] = a[i]['smiles']
		elif comp_name[i][0]=='O' or comp_name[i][0]=='H':
			 comp_smil[i] = '['+comp_name[i]+']'
		else:
			 comp_smil[i] = comp_name[i]
 
	return(comp_smil, comp_name)



'''automatically produces a module for calculating reacion rate coefficients'''
# function to generate a module for calculation of reaction rate coefficients

import datetime

def write_rate_file(reac_coef_g, reac_coef_aq, rrc, rrc_name, testf): # define function

	# inputs: ----------------------------------------------------------------------------
	# reac_coef_g - gas-phase reaction rate coefficient expression from the equation file
	# reac_coef_aq - aqueous-phase reaction rate coefficient expression from the equation file
	# rrc - expression for generic reaction rate coefficients
	# rrc_name - name given to generic reaction rate coefficients	
	# testf - flag for mode: 0 in gas-phase equation mode, 2 for test mode, 3 for
	#			aqueous-phase equation mode
	# ------------------------------------------------------------------------------------

	# open/create relevant file to write module to
	if (testf == 0):
		f = open('PyCHAM/rate_coeffs.py', mode='w')
	if (testf == 3):
		f = open('PyCHAM/rate_coeffs_aq.py', mode='w')
	if (testf == 2):
		f = open('rate_coeffs.py', mode='w')
		
	f.write('\'\'\'module for calculating reaction rate coefficients (automatically generated)\'\'\'\n')
	f.write('# module to hold expressions for calculating rate coefficients # \n') # python will convert \n to os.linesep
	f.write('# created at %s\n' %(datetime.datetime.now()))
	f.write('\n')
	f.write('import numpy\n')
	f.write('import photolysisRates\n')
	f.write('\n')

	# following part is the function (there should be an indent at the start of each line)
	# suggest using one tab
	f.write('def evaluate_rates(RO2, H2O, TEMP, lightm, time, lat, lon, act_flux_path, DayOfYear, M, N2, O2, photo_par_file, Jlen, tf):\n')
	f.write('\n')
	f.write('	# inputs: ------------------------------------------------------------------\n')
	f.write('	# RO2 - names of components included in peroxy radical list\n')	
	f.write('	# M - third body concentration (molecules/cc (air))\n')
	f.write('	# N2 - nitrogen concentration (molecules/cc (air))\n')
	f.write('	# O2 - oxygen concentration (molecules/cc (air))\n')
	f.write('	# H2O, TEMP: given by the user\n')
	f.write('	# lightm: given by the user and is 0 for lights off and 1 for on\n')
	f.write('	# reaction rate coefficients and their names parsed in eqn_parser.py \n')
	f.write('	# Jlen - number of photolysis reactions\n')
	f.write('	# tf - sunlight transmission factor\n')
	f.write('	# ------------------------------------------------------------------------\n')
	f.write('\n')
	f.write('	# calculate generic reaction rate coefficients given by chemical scheme\n')
	# code to calculate rate coefficients given by chemical scheme file
	for line in rrc:
		f.write('	%s \n' %line)
	f.write('\n')
	f.write('	# estimate and append photolysis rates\n')
	f.write('	J = photolysisRates.PhotolysisCalculation(time, lat, lon, TEMP, act_flux_path, DayOfYear, photo_par_file, Jlen, tf)\n')
	f.write('\n')
	f.write('	if lightm == 0:\n')
	f.write('		J = [0]*len(J)\n')

	# calculate the rate coefficient for each equation
	f.write('	rate_values = numpy.zeros((%i))\n' %(len(reac_coef_g)+len(reac_coef_aq)))
	# BE NOTIFIED!!!: before writing the script, 'reac_coef' must be converted to 
	# python-compatible format
	f.write('	\n')	
	f.write('	# reac_coef has been formatted so that python can recognize it\n')
	f.write('	# gas-phase reactions\n')
	for eqn_key in range (len(reac_coef_g)):
		f.write('	rate_values[%s] = %s\n' %(eqn_key, reac_coef_g[eqn_key]))
	f.write('	\n')
	f.write('	# aqueous-phase reactions\n')
	for eqn_key_aq in range (1, len(reac_coef_aq)+1):
		f.write('	rate_values[%s] = %s\n' %(eqn_key+eqn_key_aq, reac_coef_aq[eqn_key_aq-1]))
	f.write('	\n')
	f.write('	return(rate_values)\n')
	f.close()

	return()



'''module to create number size distributions'''
# Module to create log-normal size distribution. Code copied/modified from            
# http://all-geo.org/volcan01010/2013/09/how-to-use-lognormal-distributions-in-python/

from scipy import stats # import the scipy.stats module
import numpy as np
import matplotlib.pyplot as plt

def lognormal(num_bins, pmode, pconc, std, lowersize, uppersize, loc, scale, space_mode):

	# inputs: -------------------------
	
	# num_bins - number of size bins (not including wall)
	# pmode - whether particle number concentrations given in modes or explicitly
	# pconc - starting number concentration of particles (# particle/cc (air)), if this
	# is scalar, the number concentration will be split between size bins, and if it has
	# the same length as number of size bins, each element will be allocated to its
	# corresponding size bin
	# loc - shift of lognormal probability distribution function for seed particles 
	# number-size distribution (um)
	# scale - scaling factor of lognormal probability distribution function for seed 
	# particles (dimensionless) 
	# std - geometric standard deviation (dimensionless)
	# lowersize - lower radius of particles (um)
	# uppersize - upper radius of particles (um)
	# space_mode - string saying whether to space size bins logarithmically or linearly
	# ---------------------------------
    
	# volume in largest and smallest size bin (um3)
	vNb = (4.0/3.0)*np.pi*uppersize**3.0
	v1 = (4.0/3.0)*np.pi*lowersize**3.0
	
	# constant volume ratio method --------------
	# constant volume ratio (13.3 Jacobson 2005)
# 		Vrat = (vNb/v1)**(1/(num_bins-1))
	# volume at centre of each size bin (um3) (13.2 Jacobson 2005)
# 		Varr = v1*(Vrat**np.arange(0, num_bins))
# 		Varr = np.append(Varr, ((4.0*np.pi)/3.0)*(1.0**3.0)) # wall volume (um3)
# 		# radius at centre of each size bin (um)
# 		x_output = ((3.0/(4.0*np.pi))*Varr[0:-1])**(1.0/3.0)
	# ---------------------------------------
	# logarithmic method
	if space_mode == 'log':
		rad_bounds = 10.0**(np.linspace(np.log10(lowersize), 
						np.log10(uppersize), num=(num_bins+1)))
		rwid = (rad_bounds[1::]-rad_bounds[0:-1]) # width of size bins (um)
		x_output = rad_bounds[0:-1]+rwid/2.0 # particle radius (um)
		
	# ---------------------------------------
	# linear method
	if space_mode == 'lin' or space_mode == 'none':
		rad_bounds = np.linspace(lowersize, uppersize, (num_bins+1))
		# width of size bins (um)
		rwid = np.array((rad_bounds[1]-rad_bounds[0])).reshape(1)
		x_output = rad_bounds[0:-1]+rwid/2.0 # particle radius (um)

	# ---------------------------------------
	# enhance upper radius bound (um) to reduce possibility of particles growing beyond 
	# this (reversed in saving.py)
	upper_bin_rad_amp = 1.0e6
	rad_bounds[-1] = rad_bounds[-1]*upper_bin_rad_amp
	
	if (pmode == 0): # calculating the number size distribution from modes
		# number concentrations for all size bins
		Nperbin = np.zeros((num_bins))
		for i in range(len(pconc)):
			# number fraction-size distribution - enforce high resolution to ensure size
			# distribution of seed particles fully captured
			# note dividing rwid[0, 0] by 2.1 rather than 2.0 prevents the lower bound reaching 
			# zero and still gives a useful range
			# high resolution size bin radii (um)
			hires = 10**(np.linspace(np.log10(x_output[0]-rwid[0]/2.1), np.log10(uppersize), int(num_bins*1.e2)))
			# probability distribution function
			if (len(std[0]) > 1): # extract array from list
				std = std[0]
			if (len(scale[0]) > 1): # extract array from list
				scale = scale[0]
			pdf_output = stats.lognorm.pdf(hires, std[i], loc, scale[i])
			# remove any excess dimensions
			pdf_output = np.squeeze(pdf_output)
			# probability distribution function scaled to actual size bin radii
			pdf_out = np.interp(x_output, hires, pdf_output)	
			# contribute the number concentration of all size bins in 
			# this mode (# particle/cc (air))
			Nperbin += (pdf_out/sum(pdf_out))*pconc[i]
	
	# if number concentration (#/cc (air)) explicitly stated in inputs
	if (pmode == 1):
		Nperbin = np.array((pconc))
	
	Nperbin = Nperbin.reshape(-1, 1) # ensure correct shape
	
	# volume of single particles per size bin (um3) - use with lognormal method
	Varr = ((4.0*np.pi)/3.0)*(x_output**3.0)
	
	# volume bounds (um3)
	V_bounds = ((4.0*np.pi)/3.0)*(rad_bounds**3.0)


	# return number of particles per size bin (the number size distribution)
	return(Nperbin, x_output, rad_bounds, V_bounds, Varr, upper_bin_rad_amp)



'''generate the module to solve ODEs'''
# writes a module based on the supplied chemical scheme and user inputs,
# since the solved ODEs can represent gas-phase photochemistry,
# gas-particle partitioning and gas-wall partitioning, note use %f when
# writing floats from inputs

import datetime

# function to generate the ordinary differential equation (ODE)
# solver file
def ode_gen(con_infl_indx, int_tol, rowvals, wall_on, num_comp, 
		num_asb, con_C_indx, testf, eqn_num):
	
	# inputs: ------------------------------------------------
	# con_infl_indx - indices of components with constant influx
	# int_tol - integration tolerances
	# rowvals - indices of rows for Jacobian
	# wall_on - marker for whether to consider wall 
	# 	partitioning
	# num_comp - number of components
	# num_asb - number of actual size bins (excluding wall)
	# con_C_indx - index of components with constant gas-phase 
	#	concentration
	# testf - marker for whether in test mode or not
	# eqn_num - number of gas- and particle-phase reactions
	# -------------------------------------------------------
	
	# create new  file to store solver module
	f = open('PyCHAM/ode_solv.py', mode='w')
	f.write('\'\'\'solution of ODEs, generated by eqn_pars.py\'\'\'\n')
	f.write('# module to solve system of ordinary differential equations (ODEs) using solve_ivp of Scipy \n')
	f.write('# File Created at %s\n' %(datetime.datetime.now()))	
	f.write('\n')
	f.write('import numpy as np\n')
	f.write('import scipy.sparse as SP\n')
	f.write('from scipy.integrate import solve_ivp\n')
	f.write('\n')	
	f.write('# define function\n')
	f.write('def ode_solv(y, integ_step, rindx, pindx, rstoi, pstoi, \n')
	f.write('	nreac, nprod, rrc, jac_stoi, njac, jac_den_indx, \n')
	f.write('	jac_indx, Cinfl_now, y_arr, y_rind, uni_y_rind, \n')
	f.write('	y_pind, uni_y_pind, reac_col, prod_col, \n')
	f.write('	rstoi_flat, pstoi_flat, rr_arr, rr_arr_p,\n') 
	f.write('	rowvals, colptrs, num_comp, num_sb,\n')
	f.write('	wall_on, Psat, Cw, act_coeff, kw, jac_wall_indx,\n') 
	f.write('	seedi, core_diss, kelv_fac, kimt, num_asb,\n')
	f.write('	jac_part_indx,\n')
	f.write('	rindx_aq, pindx_aq, rstoi_aq, pstoi_aq,\n')
	f.write('	nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, jac_den_indx_aq, jac_indx_aq,\n')
	f.write('	y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, uni_y_pind_aq,\n')
	f.write('	reac_col_aq, prod_col_aq, rstoi_flat_aq,\n')
	f.write('	pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num, jac_mod_len,\n')
	f.write('	jac_part_hmf_indx, rw_indx, N_perbin, jac_part_H2O_indx, H2Oi):\n')
	f.write('\n')
	f.write('	# inputs: -------------------------------------\n')
	f.write('	# y - initial concentrations (moleucles/cm3)\n')
	f.write('	# integ_step - the maximum integration time step (s)\n')
	f.write('	# rindx - index of reactants per equation\n')
	f.write('	# pindx - index of products per equation\n')
	f.write('	# rstoi - stoichiometry of reactants\n')
	f.write('	# pstoi - stoichiometry of products\n')
	f.write('	# nreac - number of reactants per equation\n')
	f.write('	# nprod - number of products per equation\n')
	f.write('	# rrc - reaction rate coefficient\n')
	f.write('	# jac_stoi - stoichiometries relevant to Jacobian\n')
	f.write('	# njac - number of Jacobian elements affected per equation\n')
	f.write('	# jac_den_indx - index of component denominators for Jacobian\n')
	f.write('	# jac_indx - index of Jacobian to place elements per equation (rows)\n')
	f.write('	# Cinfl_now - influx of components with constant influx \n')
	f.write('	#		(molecules/cc/s)\n')
	f.write('	# y_arr - index for matrix used to arrange concentrations, \n')
	f.write('	#	enabling calculation of reaction rate coefficients \n')
	f.write('	# y_rind - index of y relating to reactants for reaction rate \n')
	f.write('	# 	coefficient equation\n')
	f.write('	# uni_y_rind - unique index of reactants \n')
	f.write('	# y_pind - index of y relating to products\n')
	f.write('	# uni_y_pind - unique index of products \n')
	f.write('	# reac_col - column indices for sparse matrix of reaction losses\n')
	f.write('	# prod_col - column indices for sparse matrix of production gains\n')
	f.write('	# rstoi_flat - 1D array of reactant stoichiometries per equation\n')
	f.write('	# pstoi_flat - 1D array of product stoichiometries per equation\n')
	f.write('	# rr_arr - index for reaction rates to allow reactant loss\n')
	f.write('	# 	calculation\n')
	f.write('	# rr_arr_p - index for reaction rates to allow reactant loss\n')
	f.write('	# 	calculation\n')
	f.write('	# rowvals - row indices of Jacobian elements\n')
	f.write('	# colptrs - indices of  rowvals corresponding to each column of the\n') 
	f.write('	# 	Jacobian\n')
	f.write('	# num_comp - number of components\n')
	f.write('	# num_sb - number of size bins\n')
	f.write('	# wall_on - flag saying whether to include wall partitioning\n')
	f.write('	# Psat - pure component saturation vapour pressures (molecules/cc)\n')
	f.write('	# Cw - effective absorbing mass concentration of wall (molecules/cc) \n')
	f.write('	# act_coeff - activity coefficient of components\n')
	f.write('	# kw - mass transfer coefficient to wall (/s)\n')
	f.write('	# jac_wall_indx - index of inputs to Jacobian by wall partitioning\n')
	f.write('	# seedi - index of seed material\n')
	f.write('	# core_diss - dissociation constant of seed material\n')
	f.write('	# kelv_fac - kelvin factor for particles\n')
	f.write('	# kimt - mass transfer coefficient for gas-particle partitioning (s)\n')
	f.write('	# num_asb - number of actual size bins (excluding wall)\n')
	f.write('	# jac_part_indx - index for sparse Jacobian for particle influence \n')
	f.write('	# eqn_num - number of gas- and aqueous-phase reactions \n')
	f.write('	# jac_mod_len - modification length due to high fraction of component(s)\n')
	f.write('	# 	in particle phase\n')
	f.write('	# jac_part_hmf_indx - index of Jacobian affected by water\n')
	f.write('	#	 in the particle phase\n')
	f.write('	# rw_indx - indices of rows affected by water in particle phase\n')
	f.write('	# N_perbin - number concentration of particles per size bin (#/cc)\n')
	f.write('	# jac_part_H2O_indx - sparse Jacobian indices for the effect of\n')
	f.write('	#	particle-phase water on all other components\n')
	f.write('	# H2Oi - index for water\n')
	f.write('	# ---------------------------------------------\n')
	f.write('\n')
	
	# the module if needed for testing
	if (testf > 0):
		f.write('	# gas-particle partitioning-----------------\n')
		f.write('	# transform particle phase concentrations into\n')
		f.write('	# size bins in rows, components in columns\n')
		f.write('	ymat = (y[num_comp:num_comp*(num_asb+1), 0]).reshape(num_asb, num_comp)\n')
		f.write('	# total particle-phase concentration per size bin (molecules/cc (air))\n')
		f.write('	csum = ((ymat.sum(axis=1)-ymat[:, seedi])+((ymat[:, seedi]*core_diss))).reshape(-1, 1)\n')
		f.write('	# size bins with contents \n')
		f.write('	isb = (csum[:, 0]>0.)\n')
		f.write('	\n')
		f.write('	# container for gas-phase concentrations at particle surface\n')
		f.write('	Csit = np.zeros((num_asb, num_comp))\n')
		f.write('	# mole fraction of components at particle surface\n')
		f.write('	Csit[isb, :] = (ymat[isb, :]/csum[isb, :])\n')
		f.write('	\n')
		f.write('	return(Csit)\n')
		f.close() # close file
		return()


	# testing with 16 size bins and the MCM alpha-pinene chemical scheme
	# showed that using the vectorised Python code gave just 1 %
	# increase in wall clock time compared to using numba, and won't
	# give the gradual slow down in computation that arises with numba
	# when many integration time steps are set, furthermore it means that 
	# for fast integration systems, time isn't wasted on compilation, 
	# therefore we use the
	# vectorised form by default and the numba version is commented out 
	# below
	
	f.write('	def dydt(t, y): # define the ODE(s)\n')
	f.write('		\n')
	f.write('		# inputs: ----------------\n')
	f.write('		# y - concentrations (molecules/cc), note when using scipy integrator solve_ivp, this should have shape (number of elements, 1)\n')
	f.write('		# t - time interval to integrate over (s)\n')
	f.write('		# ---------------------------------------------\n')
	f.write('		\n')
	f.write('		# ensure y is correct shape\n')
	f.write('		if (y.shape[1] > 1):\n')
	f.write('			y = y[:, 0].reshape(-1, 1)\n')
	f.write('		# empty array to hold rate of change per component\n')
	f.write('		dd = np.zeros((y.shape[0], 1))\n')
	f.write('		\n')
	
	
	if (eqn_num[0] > 0): # if gas-phase reactions present
		f.write('		# gas-phase reactions -------------------------\n')
		f.write('		# empty array to hold relevant concentrations for\n')
		f.write('		# reaction rate coefficient calculation\n')
		f.write('		rrc_y = np.ones((rindx.shape[0]*rindx.shape[1]))\n')
		f.write('		rrc_y[y_arr] = y[y_rind, 0]\n')
		f.write('		rrc_y = rrc_y.reshape(rindx.shape[0], rindx.shape[1], order = \'C\')\n')
		f.write('		# reaction rate (molecules/cc/s) \n')
		f.write('		rr = rrc[0:rindx.shape[0]]*((rrc_y**rstoi).prod(axis=1))\n')
		f.write('		# loss of reactants\n')
		f.write('		data = rr[rr_arr]*rstoi_flat # prepare loss values\n')
		f.write('		# convert to sparse matrix\n')
		f.write('		loss = SP.csc_matrix((data, y_rind, reac_col))\n')
		f.write('		# register loss of reactants\n')
		f.write('		dd[uni_y_rind, 0] -= np.array((loss.sum(axis = 1))[uni_y_rind])[:, 0]\n')
		f.write('		# gain of products\n')
		f.write('		data = rr[rr_arr_p]*pstoi_flat # prepare loss values\n')
		f.write('		# convert to sparse matrix\n')
		f.write('		loss = SP.csc_matrix((data, y_pind, prod_col))\n')
		f.write('		# register gain of products\n')
		f.write('		dd[uni_y_pind, 0] += np.array((loss.sum(axis = 1))[uni_y_pind])[:, 0]\n')
		f.write('		\n')
		
	if (eqn_num[1] > 0): # if particle-phase reactions present
		f.write('		# particle-phase reactions -------------------------\n')
		f.write('		\n')
		f.write('		# empty array to hold relevant concentrations for\n')
		f.write('		# reaction rate coefficient calculation\n')
		f.write('		# tile aqueous-phase reaction rate coefficients\n')
		f.write('		rr_aq = np.tile(rrc[rindx.shape[0]::], num_asb)\n')
		f.write('		# prepare aqueous-phase concentrations\n')
		f.write('		rrc_y = np.ones((rindx_aq.shape[0]*rindx_aq.shape[1]))\n')
		f.write('		rrc_y[y_arr_aq] = y[y_rind_aq, 0]\n')
		f.write('		rrc_y = rrc_y.reshape(rindx_aq.shape[0], rindx_aq.shape[1], order = \'C\')\n')
		f.write('		# reaction rate (molecules/cc/s) \n')
		f.write('		rr = rr_aq*((rrc_y**rstoi_aq).prod(axis=1))\n')
		f.write('		# loss of reactants\n')
		f.write('		data = rr[rr_arr_aq]*rstoi_flat_aq # prepare loss values\n')
		f.write('		# convert to sparse matrix\n')
		f.write('		loss = SP.csc_matrix((data[0, :], y_rind_aq, reac_col_aq))\n')
		f.write('		# register loss of reactants\n')
		f.write('		dd[uni_y_rind_aq, 0] -= np.array((loss.sum(axis = 1))[uni_y_rind_aq])[:, 0]\n')
		f.write('		# gain of products\n')
		f.write('		data = rr[rr_arr_p_aq]*pstoi_flat_aq # prepare loss values\n')
		f.write('		# convert to sparse matrix\n')
		f.write('		loss = SP.csc_matrix((data[0, :], y_pind_aq, prod_col_aq))\n')
		f.write('		# register gain of products\n')
		f.write('		dd[uni_y_pind_aq, 0] += np.array((loss.sum(axis = 1))[uni_y_pind_aq])[:, 0]\n')
		f.write('		\n')
	
	if (len(con_infl_indx) == 1): # if a component has a continuous gas-phase influx
		
		f.write('		# account for components with constant influx\n')	
		f.write('		dd[[')	
		for Ci in range(len(con_infl_indx)):
			# components prior to the last in the continuous influx group
			if (Ci < len(con_infl_indx)-1):
				f.write('%d, ' %int(con_infl_indx[Ci]))
			else: # last component in the continuous influx group
				f.write('%d], 0] += Cinfl_now[:, 0]\n' %int(con_infl_indx[Ci]))
		f.write('		\n')
		
	if (len(con_C_indx) > 0): # if any components have a constant gas-phase concentration
		f.write('		dd[[')
		for Ci in range(len(con_C_indx)):
			if Ci<len(con_C_indx)-1:
				f.write('%d, ' %int(con_C_indx[Ci]))
			else:
				f.write('%d], 0] = 0. \n' %int(con_C_indx[Ci]))
		f.write('		\n')
		
	# note the following needs two indents (as for the reaction section), so that it
	# sits within the dydt function
	if (num_asb > 0): # include gas-particle partitioning in ode solver
		f.write('		# gas-particle partitioning-----------------\n')
		f.write('		# transform particle phase concentrations into\n')
		f.write('		# size bins in rows, components in columns\n')
		f.write('		ymat = (y[num_comp:num_comp*(num_asb+1), 0]).reshape(num_asb, num_comp)\n')
		f.write('		# force all components in size bins with no particle to zero\n')
		f.write('		ymat[N_perbin[:, 0] == 0, :] = 0\n')	
		f.write('		# total particle-phase concentration per size bin (molecules/cc (air))\n')
		f.write('		csum = ((ymat.sum(axis=1)-ymat[:, seedi].sum(axis=1))+((ymat[:, seedi]*core_diss).sum(axis=1)).reshape(-1)).reshape(-1, 1)\n')
		f.write('		# size bins with contents \n')
		f.write('		isb = (csum[:, 0] > 0.)\n')
		f.write('		\n')
		f.write('		if (any(isb)): # if particle-phase components present\n')
		f.write('			# container for gas-phase concentrations at particle surface\n')
		f.write('			Csit = np.zeros((num_asb, num_comp))\n')
		f.write('			# mole fraction of components at particle surface\n')
		f.write('			Csit[isb, :] = (ymat[isb, :]/csum[isb, :])\n')	
		f.write('			# gas-phase concentration of components at\n')
		f.write('			# particle surface (molecules/cc (air))\n')
		f.write('			Csit[isb, :] = Csit[isb, :]*Psat[isb, :]*kelv_fac[isb]*act_coeff[isb, :]\n')	
		f.write('			# partitioning rate (molecules/cc/s)\n')
		f.write('			dd_all = kimt*(y[0:num_comp, 0].reshape(1, -1)-Csit)\n')
		f.write('			# gas-phase change\n')
		f.write('			dd[0:num_comp, 0] -= dd_all.sum(axis=0)\n')
		f.write('			# particle change\n')
		f.write('			dd[num_comp:num_comp*(num_asb+1), 0] += (dd_all.flatten())\n')
		f.write('		\n')
		
	if (wall_on > 0): # include gas-wall partitioning in ode solver
		f.write('		# gas-wall partitioning ----------------\n')
		f.write('		# concentration on wall (molecules/cc (air))\n')
		f.write('		Csit = y[num_comp*(num_asb+1):num_comp*(num_asb+2), 0]\n')
		f.write('		# saturation vapour pressure on wall (molecules/cc (air))\n')
		f.write('		# note, just using the top rows of Psat and act_coeff\n')
		f.write('		# as do not need the repetitions over size bins\n')
		f.write('		if (Cw > 0.):\n')
		f.write('			Csit = Psat[0, :]*(Csit/Cw)*act_coeff[0, :]\n')
		f.write('			# rate of transfer (molecules/cc/s)\n')
		f.write('			dd_all = kw*(y[0:num_comp, 0]-Csit)\n')
		f.write('			dd[0:num_comp, 0] -= dd_all # gas-phase change\n')
		f.write('			dd[num_comp*num_sb:num_comp*(num_sb+1), 0] += dd_all # wall change\n')
		f.write('		\n')
	
	# non-vectorised code for use with numba compiler
	#	f.write('	# numba compiler to convert to machine code\n')
	#	f.write('	@jit(f8[:](f8, f8[:]), nopython=True, cache=False)\n')
	#	f.write('	# ode solver -------------------------------------\n')
	#	f.write('	def dydt(t, y): # define the ODE(s)\n')
	#	f.write('		\n')
	#	f.write('		# empty array to hold rate of change per component\n')
	#	f.write('		dd = np.zeros((len(y)))\n')
	#	f.write('		# gas-phase rate of change -----------------------\n')
	#	f.write('		for i in range(nreac.shape[0]): # equation loop\n')
	#	f.write('			# gas-phase rate of change (molecules/cc (air).s)\n')
	#	f.write('			if (y[rindx[i, 0:nreac[i]], 0]==0.0).sum()>0:\n')	
	#	f.write('				continue # if any reactants not present skip this reaction\n')			
	#	f.write('			else:\n')
	#	f.write('				gprate = ((y[rindx[i, 0:nreac[i]], 0]**rstoi[i, 0:nreac[i]]).prod())*rrc[i]\n')
	#	f.write('				# loss of reactants\n')
	#	f.write('				dd[rindx[i, 0:nreac[i]]] -= gprate*rstoi[i, 0:nreac[i]]\n')
	#	f.write('				# gain of products\n')
	#	f.write('				dd[pindx[i, 0:nprod[i]]] += gprate*pstoi[i, 0:nprod[i]]\n')
	#	f.write('		\n')
	#	# only write next section if particles present
	#	if (num_asb > 0):
	#		f.write('		# if size bins present estimate gas-particle partitioning\n')
	#		f.write('		for ibin in range(num_asb): # size bin loop\n')
	#		f.write('			# particle-phase concentrations in this size bin (moclecules/cc)\n')
	#		f.write('			Csit = y[num_comp*(ibin+1):num_comp*(ibin+2), 0]\n')
	#		f.write('			# prepare for # sum of molecular concentrations per bin (molecules/cc (air))\n')
	#		f.write('			conc_sum = np.zeros((1))\n')
	#		f.write('			conc_sum[0] = ((Csit.sum()-Csit[seedi])+Csit[seedi]*core_diss)\n')
	#		f.write('			# only need to continue if particles present\n')
	#		f.write('			if (conc_sum[0]>1.e-20):\n')
	#		f.write('				# particle surface gas-phase concentration (molecules/cc (air))\n')
	#		f.write('				Csit = (Csit/conc_sum)*Psat[0, :]*kelv_fac[ibin]*act_coeff[0, :]\n')
	#		f.write('				# partitioning rate (molecules/cc.s)\n')
	#		f.write('				dydt_all = kimt[ibin, :]*(y[0:num_comp, 0]-Csit)\n')
	#		f.write('				dd[0:num_comp] -= dydt_all # gas-phase change\n')
	#		f.write('				# particle-phase change\n')
	#		f.write('				dd[num_comp*(ibin+1):num_comp*(ibin+2)] += dydt_all\n')
	#		f.write('		\n')
	#	# only write next section if gas-wall partitioning active
	#	if (wall_on==1):
	#		f.write('		if (Cw > 0.): # only consider if wall present\n')
	#		f.write('			# if wall consideration turned on, estimate gas-wall partitioning\n')
	#		f.write('			# concentration at wall (molecules/cc (air))\n')
	#		f.write('			Csit = y[num_comp*num_sb:num_comp*(num_sb+1), 0]\n')
	#		f.write('			Csit = (Psat[0, :]*(Csit/Cw)*act_coeff[0, :]) # with Raoult term\n')
	#		f.write('			dydt_all = (kw)*(y[0:num_comp, 0]-Csit)\n')
	#		f.write('			dd[0:num_comp] -= dydt_all # gas-phase change\n')
	#		f.write('			# wall concentration change \n')
	#		f.write('			dd[num_comp*num_sb:num_comp*(num_sb+1)] += dydt_all\n')
	#		f.write('			\n')
	f.write('		\n')
	
	f.write('		dd = (dd[:, 0]).reshape(num_sb+1, num_comp)\n')
	f.write('		# force all components in size bins with no particle to zero\n')
	f.write('		if (num_asb > 0):\n')
	f.write('			dd[1:num_asb+1, :][N_perbin[:, 0] == 0, :] = 0\n')
	f.write('		# return to array, note that consistent with the solve_ivp manual, this ensures dd is\n')
	f.write('		# a vector rather than matrix, since y0 is a vector\n')
	f.write('		dd = dd.flatten()\n')
	f.write('		return (dd)\n')
	f.write('\n')
	
	# testing with 16 size bins and the MCM alpha-pinene chemical scheme
	# showed that defining the Jacobian gave a 10 % reduction in processing
	# time, therefore this set by default
	f.write('	def jac(t, y): # define the Jacobian\n')
	f.write('		\n')
	f.write('		# inputs: ----------------\n')
	f.write('		# y - concentrations (molecules/cc), note when using scipy integrator solve_ivp, this should have shape (number of elements, 1)\n')
	f.write('		# t - time interval to integrate over (s)\n')
	f.write('		# ---------------------------------------------\n')
	f.write('		\n')
	f.write('		# ensure y is correct shape\n')
	f.write('		if (y.ndim == 2):\n')
	f.write('			if (y.shape[1] > 1):\n')
	f.write('				y = y[:, 0].reshape(-1, 1)\n')
	f.write('		if (y.ndim <= 1):\n')
	f.write('			y = y.reshape(-1, 1)\n')
	f.write('		\n')
	
	f.write('		# elements of sparse Jacobian matrix\n')
	if (num_asb > 0): # include any particle-phase modifiers
		f.write('		data = np.zeros((%s+jac_mod_len))\n' %len(rowvals))
	else: # don't include any particle-phase modifiers
		f.write('		data = np.zeros((%s))\n' %len(rowvals))
	f.write('		\n')
	
	if (eqn_num[0] > 0): # if gas-phase reactions present
		f.write('		for i in range(rindx.shape[0]): # gas-phase reaction loop\n')
		f.write('			# reaction rate (molecules/cc/s)\n')
		f.write('			rr = rrc[i]*(y[rindx[i, 0:nreac[i]], 0].prod())\n')
		f.write('			# prepare Jacobian inputs\n')
		f.write('			jac_coeff = np.zeros((njac[i, 0]))\n')
		f.write('			# only fill Jacobian if reaction rate sufficient\n')
		f.write('			if (rr != 0.):\n')
		f.write('				jac_coeff = (rr*(jac_stoi[i, 0:njac[i, 0]])/\n')
		f.write('				(y[jac_den_indx[i, 0:njac[i, 0]], 0]))\n')
		f.write('			data[jac_indx[i, 0:njac[i, 0]]] += jac_coeff\n')
		f.write('		\n')
	
	if (eqn_num[1] > 0): # if particle-phase reactions present
		f.write('		n_aqr = nreac_aq.shape[0] # number of aqueous-phase reactions \n')
		f.write('		aqi = 0 # aqueous-phase reaction counter\n')
		f.write('		\n')
		f.write('		for i in range(rindx.shape[0], rrc.shape[0]): # aqueous-phase reaction loop\n')
		f.write('			# reaction rate (molecules/cc/s)\n')
		f.write('			rr = rrc[i]*(y[rindx_aq[aqi::n_aqr, 0:nreac_aq[aqi]], 0].prod(axis=1))\n')
		f.write('			# spread along affected components\n')
		f.write('			rr = rr.reshape(-1, 1)\n')
		f.write('			rr = (np.tile(rr, int(njac_aq[aqi, 0]/(num_sb-wall_on)))).flatten(order=\'C\')\n')
		f.write('			# prepare Jacobian inputs\n')
		f.write('			jac_coeff = np.zeros((njac_aq[aqi, 0]))\n')
		f.write('			nzi = (rr != 0)\n')
		f.write('			jac_coeff[nzi] = (rr[nzi]*((jac_stoi_aq[aqi, 0:njac_aq[aqi, 0]])[nzi])/\n')
		f.write('				((y[jac_den_indx_aq[aqi, 0:njac_aq[aqi, 0]], 0])[nzi]))\n')
		f.write('			# stack size bins\n')
		f.write('			jac_coeff = jac_coeff.reshape(int(num_sb-wall_on), int(njac_aq[aqi, 0]/(num_sb-wall_on)), order=\'C\')\n')
		f.write('			data[jac_indx_aq[aqi::n_aqr, 0:(int(njac_aq[aqi, 0]/(num_sb-wall_on)))]] += jac_coeff\n')
		f.write('			\n')
		f.write('			aqi += 1 # keep count on aqueous-phase reactions \n')
	f.write('		\n')
	
	if (num_asb > 0): # include gas-particle partitioning in ode solver Jacobian
		f.write('		# gas-particle partitioning\n')
		f.write('		part_eff = np.zeros((%s))\n' %((num_comp)*(num_asb+1)+((num_comp)*(num_asb*2))))
		f.write('		if (sum(N_perbin[:, 0]) > 0.): # if any particles present \n')
		f.write('			part_eff[0:%s:%s] = -kimt.sum(axis=0) # effect of gas on gas\n' %(num_comp*(num_asb+1), (num_asb+1)))
		f.write('		\n')
		f.write('		# empty array for any particle-on-gas and particle-on-particle effects on water in the particle-phase for rows of Jacobian\n')
		f.write('		part_eff_rw = np.zeros((len(jac_part_hmf_indx)))\n')
		f.write('		# empty array for any particle-on-gas and particle-on-particle effects of water in the particle-phase on non-water components in the particle-phase for columns of Jacobian\n')
		f.write('		part_eff_cl = np.zeros((len(jac_part_H2O_indx)))\n')
		f.write('		# starting index for jacobian row inputs for effect on water\n')
		f.write('		sti_rw = 0 \n')
		f.write('		\n')
		f.write('		# transform particle phase concentrations into\n')
		f.write('		# size bins in rows, components in columns\n')
		f.write('		ymat = (y[num_comp:num_comp*(num_asb+1), 0]).reshape(num_asb, num_comp)\n')
		f.write('		ymat[N_perbin[:, 0] == 0, :] = 0 # ensure zero components where zero particles\n')
		f.write('		# total particle-phase concentration per size bin (molecules/cc (air))\n')
		f.write('		csum = ymat.sum(axis=1)-ymat[:, seedi].sum(axis=1)+(ymat[:, seedi]*core_diss).sum(axis=1)\n')
		f.write('		\n')
		f.write('		# effect of particle on gas\n')
		f.write('		for isb in range(int(num_asb)): # size bin loop\n')
		f.write('			if (csum[isb] > 0): # if components present in this size bin\n')
		f.write('				# effect of gas on particle\n')
		f.write('				part_eff[1+isb:num_comp*(num_asb+1):num_asb+1] = +kimt[isb, :]\n')
		f.write('				# start index\n')
		f.write('				sti = int((num_asb+1)*num_comp+isb*(num_comp*2))\n')
		f.write('				# diagonal index\n')
		f.write('				diag_indxg = sti+np.arange(0, num_comp*2, 2).astype(\'int\')\n')
		f.write('				diag_indxp = sti+np.arange(1, num_comp*2, 2).astype(\'int\')\n')
		f.write('				# prepare for diagonal (component effect on itself)\n')
		f.write('				diag = kimt[isb, :]*Psat[0, :]*act_coeff[0, :]*kelv_fac[isb, 0]*(-(csum[isb]-ymat[isb, :])/(csum[isb]**2.)) \n')
		f.write('				# implement to part_eff\n')
		f.write('				part_eff[diag_indxg] -= diag\n')
		f.write('				part_eff[diag_indxp] += diag\n')
		f.write('				\n')
		f.write('				if (rw_indx[isb] > -1): # if water in this size bin \n') 
		f.write('					# prepare for row(s) (particle-phase non-water component effects on water in particle phase)\n')
		f.write('					rw = kimt[isb, rw_indx[isb]]*Psat[0, rw_indx[isb]]*act_coeff[0, rw_indx[isb]]*kelv_fac[isb, 0]*(-(-ymat[isb, rw_indx[isb]])/(csum[isb]**2.)) \n')
		f.write('					# indices\n')
		f.write('					indxg = sti_rw+np.arange(0, ((num_comp-1)*2), 2).astype(\'int\')\n')
		f.write('					indxp = sti_rw+np.arange(1, ((num_comp-1)*2), 2).astype(\'int\')\n')
		f.write('					# implement to part_eff_rw\n')
		f.write('					part_eff_rw[indxg] -= rw\n')
		f.write('					part_eff_rw[indxp] += rw\n')
		f.write('					\n')
		f.write('					# prepare for column(s) (particle-phase water effect on non-water in particle phase)\n')
		f.write('					#cl = kimt[isb, :]*Psat[0, :]*act_coeff[0, :]*kelv_fac[isb, 0]*(-(-ymat[isb, :])/(csum[isb]**2.))\n')
		f.write('					#cl = np.zeros((num_comp))\n')
		f.write('					# remove water\n')
		f.write('					#cl = np.concatenate((cl[0:H2Oi], cl[H2Oi+1::]))\n')
		f.write('					#indxg = sti_rw+np.arange(0, (num_comp-1)).astype(\'int\')\n')
		f.write('					#indxp = sti_rw+np.arange((num_comp-1), (num_comp-1)*2).astype(\'int\')\n')
		f.write('					# implement to part_eff_cl\n')
		f.write('					#part_eff_cl[indxg] -= cl\n')
		f.write('					#part_eff_cl[indxp] += cl\n')
		f.write('					\n')
		f.write('					# starting index update\n')
		f.write('					sti_rw += (num_comp-1)*2\n')
		f.write('		\n')
		f.write('		data[jac_part_indx] += part_eff # diagonal\n')
		f.write('		data[jac_part_hmf_indx] += part_eff_rw # rows\n')
		f.write('		#data[jac_part_H2O_indx] += part_eff_cl # columns\n')
		f.write('		\n')
		
	if (wall_on > 0): # include gas-wall partitioning in ode solver Jacobian
		f.write('		if (Cw > 0.):\n')
		f.write('			wall_eff = np.zeros((%s))\n' %(num_comp*4))
		f.write('			wall_eff[0:%s:2] = -kw # effect of gas on gas \n' %(num_comp*2))
		f.write('			wall_eff[1:%s:2] = +kw # effect of gas on wall \n' %(num_comp*2))
		f.write('			# effect of wall on gas\n')
		f.write('			wall_eff[%s:%s:2] = +kw*(Psat[0,:]*act_coeff[0, :]/Cw) \n' %(num_comp*2, num_comp*4))
		f.write('			# effect of wall on wall\n')
		f.write('			wall_eff[%s+1:%s:2] = -kw*(Psat[0,:]*act_coeff[0, :]/Cw) \n' %(num_comp*2, num_comp*4))
		f.write('			data[jac_wall_indx] += wall_eff\n')
		f.write('		\n')

	f.write('		# create Jacobian\n')
	f.write('		j = SP.csc_matrix((data, rowvals, colptrs))\n')
	f.write('		\n')
	f.write('		return(j)\n')
	f.write('	\n')
	f.write('	# set ODE solver tolerances\n')
	f.write('	atol = %s\n'%int_tol[0])
	f.write('	rtol = %s\n'%int_tol[1])
	f.write('	\n')
	f.write('	# call on the ODE solver, note y contains the initial condition(s) (molecules/cc (air)) and must be 1D even though y in dydt and jac has shape (number of elements, 1)\n')
	f.write('	sol = solve_ivp(dydt, [0, integ_step], y, atol = atol, rtol = rtol, method = \'BDF\', t_eval = [integ_step], vectorized = True, jac = jac)\n')
	f.write('	\n')
	f.write('	# force all components in size bins with no particle to zero\n')
	f.write('	y = np.squeeze(sol.y)\n')
	f.write('	y = y.reshape(num_sb+1, num_comp)\n')
	f.write('	if (num_asb > 0):\n')
	f.write('		y[1:num_asb+1, :][N_perbin[:, 0] == 0, :] = 0\n')
	f.write('	# return to array\n')
	f.write('	y = y.flatten()\n')
	f.write('	\n')
	f.write('	# return concentration(s) and time(s) following integration\n')
	f.write('	return(y, sol.t)\n')
	f.close() # close file




'''default model variables'''
# the default model variables inputs

import numpy as np
import os
import pickle # used for storing values

def def_mod_var(caller): # define function

	# inputs: -----------------------------------------------
	# caller - mark for calling function
	# -------------------------------------------------------

	# names ---------------------------------------------------------------------------------
	# name of folder to save results to
	sav_nam = 'default_res_name'
	sch_name = os.getcwd()+'/PyCHAM/input/example_scheme.txt'
	# markers to isolate sections of chemical scheme based on MCM KPP format
	chem_sch_mark = ['%', 'RO2', '+', '', '', ';', '+', ';', '', '%', ':', ';']
	xml_name = os.getcwd()+'/PyCHAM/input/example_xml.xml' # xml file path

	# times ----------------------------------------------------------------------------------
	# time interval between updates to integration inputs (s)
	update_stp = 1.
	tot_time = 1. # total time to integrate over (s)
	save_step = 1. # time interval between saving results (s)
	
	# initial atmosphere -----------------------------------------------------------------
	temp = np.array((298.15)).reshape(1) # temperature of experiment (K)
	tempt = np.array((0.0)).reshape(1) # time that temperatures reached (s)	
	RH = 0.65 # humidity of experiment (fraction of 1)
	Press = 9.8e4 # air pressure during experiment (Pa)
	
	# particle section ----------------------------------------------------------------
	siz_stru = 0 # size structure (0 for moving-centre, 1 for full-moving)
	num_sb = 0 # number of particle size bins
	# whether particle number concentrations expressed by modes (0) or explicitly	
	pmode = 0
	# initial concentration of particles (# particles/cc (air))
	pconc = np.zeros((1, 1))
	pconct = np.zeros((1, 1)) # times of particle injection (s)
	seed_mw = 132.14 # molecular weight of seed material (g/mol)
	seed_diss = [1.] # dissociation constant of seed material
	seed_dens = 1. # density of seed material (g/cc)
	seed_name = ['core'] # name of component forming seed material
	seedVr = [1.] # volume ratio of seed components
	lowsize = 0. # smallest size bin boundary (radius) (um)
	uppsize = 5.e-1 # largest size bin boundary (radius) (um)
	space_mode = 'lin' # treatment for spacing between size bins
	std = np.ones((1, 1))*1.2 # standard deviation for particle number size distribution
	mean_rad = np.ones((1, 1))*-1.e6 # mean radius for particle number size distribution (um)
	new_partr = 2.e-7 # radius of newly nucleated particles (cm)
	# nucleation parameters
	nucv1 = 0.
	nucv2 = 0.
	nucv3 = 0.
	nuc_comp = ['core'] # chemical scheme name of nucleating component
	# marker to say whether or not to adapt integration time interval 
	# and initial condition update to nucleation
	nuc_ad = 1
	ser_H2O = 1 # whether to serialise gas-particle partitioning of water

	# component inputs ------------------------------------------------------------
	# chemical scheme name of components present initially,
	comp0 = np.array(())
	# initial concentrations (ppb),
	y0 = np.array(())	
	con_infl_nam = [] # chemical scheme names of components with continuous influx
	con_infl_t = np.array(())
	# influx rate of components with continuous influx (ppb/s)
	con_infl_C = np.array(())
	# times of component influx (s)
	const_infl_t = []	
	# chemical scheme name of components with constant concentration	
	const_comp = []
	# name(s) of component(s) injected instantaneously after start of experiment
	Compt = []
	# times at which instantaneous injection of component(s) occur after 
	# experiment start (s)
	injectt = []
	# concentration(s) (ppb) of component(s) injected instantaneously after 
	# experiment start
	Ct = np.zeros((0, 0))

	# lights -------------------------------------------------------------------------------
	light_stat = [0] # light status
	light_time = np.zeros((1, 1)) # time that light status attained (s)
	daytime = 0. # time of day experiment starts (s)
	lat = 0. # latitude of experiment (degrees)
	lon = 0. # longitude of experiment (degrees)
	af_path = 'no' # path to customised (non-MCM) actinic flux file
	# path to file containing absorption cross-sections and quantum yields
	photo_path = str(os.getcwd() + '/PyCHAM/photofiles/' + 'MCMv3.2')
	dayOfYear = 1 # number of days since 31st December that experiment conducted
	tf = 1. # transmission factor for natural sunlight (0-1 fraction)
	# marker to say whether or not to adapt integration time interval 
	# and initial condition update to changing natural light intensity
	light_ad = 1

	# deposition of particles and vapours to wall ------------------------------------------
	wall_on = 1 # marker for whether to consider wall (0 for no, 1 for yes)
	Cw = 0. # effective absorbing mass of wall (g/m3 (air))
	kw = 0. # gas-wall mass transfer coefficient (/s)

	inflectDp = 0. # diameter of deposition function inflection
	pwl_xpre = 0. # gradient before inflection
	pwl_xpro = 0. # gradient after inflection
	inflectk = 0. # rate at inflection
	chamSA = 42. # chamber surface area (m2) 
	Rader = -1 # flag for deposition to wall treatment (0 for customised, 1 for Rader and McMurry (1985))
	p_char = 0. # average number of charges per particle (/particle)
	e_field = 0. # average electric field inside chamber (g.m/A.s3)

	# miscellaneous ------------------------------------------------------------------------
	# chemical scheme name of components to track the change tendencies of	
	dydt_trak = []
	# chemical scheme names of components with densities manually set
	dens_comp = []
	# manually assigned densities (g/cc)
	dens = []
	# chemical scheme names of components with vapour pressures manually set
	vol_comp = []
	# manually assigned vapour pressures (Pa)
	volP = []
	# names of components (corresponding to chemical scheme name) with 
	# activity coefficient stated in act_user
	act_comp = []
	# user-specified activity coefficients of components with names given 
	# in act_comp
	act_user = []
	# names of components with user-defined accommodation coefficients	
	accom_comp = []
	# user-defined accommodation coefficients
	accom_val = []
	# the gas-particle partitioning cutoff (Pa)
	partit_cutoff = []

	if (caller == 0): # called from PyCHAM
		uman_up = 0 # marker for whether to update the UManSysProp folder
	if (caller == 1): # called from Travis
		uman_up = 1
	int_tol = [1.e-3, 1.e-4]
	coag_on = 1 # whether to model coagulation
	dil_fac = 0. # dilution factor

	# prepare for pickling
	list_vars = [sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, tot_time, comp0, y0, temp, tempt, RH, Press, wall_on, Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, seed_diss, seed_dens, seedVr, light_stat, light_time, daytime, lat, lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O]

	
	# path to store for variables
	input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		
	with open(input_by_sim, 'wb') as f: # the file to be used for pickling
		pickle.dump(list_vars,f) # pickle
		f.close() # close


	return(sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, tot_time, comp0, y0, temp, tempt, RH, Press, wall_on, Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, seed_diss, seed_dens, seedVr, light_stat, light_time, daytime, lat, lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O)



'''opens and returns user-defined variables from pickle file'''
# acts as the conduit between the user input process and the 
# software collecting the inputs

#import ui_check
import pickle
import os
import ui_check

# define function
def share(source):

	# inputs: ---------------------------------------------------------------------
	# source - marker for calling script
	# -----------------------------------------------------------------------------

	if (source == 0): # when called from PyCHAM script

		# path to store for variables
		input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		with open(input_by_sim, 'rb') as pk:
			[sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, 
			tot_time, comp0, y0, temp, tempt, RH, Press, wall_on,
			Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, 
			save_step, const_comp, Compt, injectt, Ct, seed_name,
			seed_mw, seed_diss, seed_dens, seedVr,
			light_stat, light_time, daytime, lat, lon, af_path, 
			dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, 
			dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, 
			accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, 
			nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, 
			inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O] = pickle.load(pk)
			pk.close()	
			
			# check on inputs
			[wall_on, pconc, lowsize, std, mean_rad, new_partr, chamR, chem_sch_mark, 
			af_path, int_tol, update_stp, tot_time, siz_str, light_stat, light_time, seedVr, 
			seed_diss, uman_up] = ui_check.ui_check(sav_nam, sch_name,
			wall_on, 0, siz_stru, num_sb, pmode, pconc, pconct, lowsize, std, mean_rad, new_partr, chamSA, 
			chem_sch_mark, af_path, int_tol, update_stp, tot_time, 
			RH, uman_up, light_stat, light_time, injectt, Ct,
			dens_comp, dens, seed_name, seedVr, seed_diss, partit_cutoff)
			
		return(sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, 
			tot_time, comp0, y0, temp, tempt, RH, Press, wall_on,
			Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, 
			save_step, const_comp, Compt, injectt, Ct, seed_name,
			seed_mw, seed_diss, seed_dens, seedVr,
			light_stat, light_time, daytime, lat, lon, af_path, 
			dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, 
			dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, 
			accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, 
			nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, 
			inflectk, chamR, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O)

	if (source == 1): # when called from plotting script
		# path to store for variables
		input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		with open(input_by_sim, 'rb') as pk:
			[sav_name, sch_name, indx_plot, Comp0] = pickle.load(pk)
			pk.close()


		return(sav_name, sch_name, indx_plot, Comp0)



'''generates module that records change tendencies'''
# function to automatically generate a module that is used to record the tendency
# of components (components specified by the user, their index given by rec_comp_index) 
# to change in response to box model 
# mechanisms - gets called inside ode_gen on each time step

import datetime

def write_dydt_rec(): # define function

	f = open('PyCHAM/dydt_rec.py', mode='w')
	f.write('\'\'\'module for calculating and recording change tendency of components\'\'\'\n')
	f.write('# changes due to gas-phase photochemistry and partitioning are included\n')
	f.write('\n')
	f.write('# File Created at %s\n' %(datetime.datetime.now()))
	f.write('\n')
	f.write('import numpy as np \n')
	f.write('\n')
	# following part is the function (there should be an indent at the start of each line)
	# suggest 1 Tab
	f.write('def dydt_rec(y, rindx, rstoi, reac_coef, pindx, pstoi, nprod, step, dydt_vst, nreac, num_sb, num_speci, pconc, core_diss, Psat, kelv_fac, kimt, kwgt, Cw, act_coeff, seedi):\n')
	f.write('	\n')
	f.write('	# loop through components to record the tendency of change \n')
	f.write('	for compi in dydt_vst.get(\'comp_index\'): \n')
	f.write('		# open relevant dictionary value \n')
	f.write('		dydt_rec = dydt_vst.get(compi) \n')
	f.write('		# keep count on relevant reactions \n')
	f.write('		reac_count = 0 \n')
	f.write('		# loop through relevant reactions \n')
	f.write('		for i in dydt_rec[0, 0:-2]: # final two rows for particle- and wall-partitioning \n')
	f.write('			i = int(i) # ensure reaction index is integer - this necessary because the dydt_rec array is float (the tendency to change records beneath its first row are float) \n')
	f.write('			# estimate gas-phase change tendency for each reaction involving this component \n')
	f.write('			gprate = ((y[rindx[i, 0:nreac[i]]]**rstoi[i, 0:nreac[i]]).prod())*reac_coef[i] \n')
	f.write('			# identify whether this component reacted and/or produced\n')
	f.write('			# and get its stoichiometric index\n')
	f.write('			stoi_indx = (np.where(rindx[i, 0:nreac[i]]==compi))[0]\n')
	f.write('			if len(stoi_indx)>0: \n')
	f.write('				dydt_rec[step+1, reac_count] -= rstoi[i, stoi_indx]*((gprate))\n')
	f.write('			stoi_indx = (np.where(pindx[i, 0:nprod[i]]==compi))[0]\n')
	f.write('			if len(stoi_indx)>0: \n')
	f.write('				dydt_rec[step+1, reac_count] += pstoi[i, stoi_indx]*((gprate))\n')
	f.write('			reac_count += 1 \n')
	f.write('		# now estimate and record tendency to change due to particle- and wall-partitioning  \n')
	f.write('		# particle-partitioning \n')
	f.write('		for ibin in range(num_sb-1): # size bin loop\n')
	f.write('			Csit = y[num_speci*(ibin+1):num_speci*(ibin+2)]\n')
	f.write('			conc_sum = np.zeros((1)) \n')
	f.write('			if any(pconc > 0.): # if seed particles present \n')
	f.write('				conc_sum[0] = ((Csit.sum()-Csit[seedi])+Csit[seedi]*core_diss)\n')
	f.write('			else: \n')
	f.write('				conc_sum[0] = Csit.sum() \n')
	f.write('			# prevent numerical error due to division by zero \n')
	f.write('			ish = conc_sum==0.0 \n')
	f.write('			conc_sum[ish] = 1.0e-40 \n')
	f.write('			# particle surface gas-phase concentration (molecules/cc (air)) \n')
	f.write('			Csit = (Csit[compi]/conc_sum)*Psat[0, compi]*kelv_fac[ibin, 0]*act_coeff[0, compi] \n')
	f.write('			# partitioning rate (molecules/cc.s) \n')
	f.write('			dydt_all = kimt[ibin, compi]*(y[compi]-Csit) \n')
	f.write('			# gas-phase change (molecules/cc/s) \n')
	f.write('			dydt_rec[step+1, reac_count] -= dydt_all \n')
	f.write('		# wall-partitioning \n')
	f.write('		if (kwgt)>1.0e-10: \n')
	f.write('			# concentration at wall (molecules/cc (air)) \n')
	f.write('			Csit = y[num_speci*num_sb:num_speci*(num_sb+1)] \n')
	f.write('			Csit = (Psat[0, :]*(Csit/Cw)*act_coeff[0, compi])\n')
	f.write('			dydt_all = (kwgt)*(y[compi]-Csit[compi]) \n')
	f.write('			# gas-phase change (molecules/cc/s) \n')
	f.write('			dydt_rec[step+1, reac_count+1] -= dydt_all \n')
	f.write('		\n')
	f.write('	return(dydt_vst) \n')
	f.close()



'''module for calculating and recording change tendency of components'''
# changes due to gas-phase photochemistry and partitioning are included

# File Created at 2020-12-15 18:13:17.839200

import numpy as np 

def dydt_rec(y, rindx, rstoi, reac_coef, pindx, pstoi, nprod, step, dydt_vst, nreac, num_sb, num_speci, pconc, core_diss, Psat, kelv_fac, kimt, kwgt, Cw, act_coeff, seedi):
	
	# loop through components to record the tendency of change 
	for compi in dydt_vst.get('comp_index'): 
		# open relevant dictionary value 
		dydt_rec = dydt_vst.get(compi) 
		# keep count on relevant reactions 
		reac_count = 0 
		# loop through relevant reactions 
		for i in dydt_rec[0, 0:-2]: # final two rows for particle- and wall-partitioning 
			i = int(i) # ensure reaction index is integer - this necessary because the dydt_rec array is float (the tendency to change records beneath its first row are float) 
			# estimate gas-phase change tendency for each reaction involving this component 
			gprate = ((y[rindx[i, 0:nreac[i]]]**rstoi[i, 0:nreac[i]]).prod())*reac_coef[i] 
			# identify whether this component reacted and/or produced
			# and get its stoichiometric index
			stoi_indx = (np.where(rindx[i, 0:nreac[i]]==compi))[0]
			if len(stoi_indx)>0: 
				dydt_rec[step+1, reac_count] -= rstoi[i, stoi_indx]*((gprate))
			stoi_indx = (np.where(pindx[i, 0:nprod[i]]==compi))[0]
			if len(stoi_indx)>0: 
				dydt_rec[step+1, reac_count] += pstoi[i, stoi_indx]*((gprate))
			reac_count += 1 
		# now estimate and record tendency to change due to particle- and wall-partitioning  
		# particle-partitioning 
		for ibin in range(num_sb-1): # size bin loop
			Csit = y[num_speci*(ibin+1):num_speci*(ibin+2)]
			conc_sum = np.zeros((1)) 
			if any(pconc > 0.): # if seed particles present 
				conc_sum[0] = ((Csit.sum()-Csit[seedi])+Csit[seedi]*core_diss)
			else: 
				conc_sum[0] = Csit.sum() 
			# prevent numerical error due to division by zero 
			ish = conc_sum==0.0 
			conc_sum[ish] = 1.0e-40 
			# particle surface gas-phase concentration (molecules/cc (air)) 
			Csit = (Csit[compi]/conc_sum)*Psat[0, compi]*kelv_fac[ibin, 0]*act_coeff[0, compi] 
			# partitioning rate (molecules/cc.s) 
			dydt_all = kimt[ibin, compi]*(y[compi]-Csit) 
			# gas-phase change (molecules/cc/s) 
			dydt_rec[step+1, reac_count] -= dydt_all 
		# wall-partitioning 
		if (kwgt)>1.0e-10: 
			# concentration at wall (molecules/cc (air)) 
			Csit = y[num_speci*num_sb:num_speci*(num_sb+1)] 
			Csit = (Psat[0, :]*(Csit/Cw)*act_coeff[0, compi])
			dydt_all = (kwgt)*(y[compi]-Csit[compi]) 
			# gas-phase change (molecules/cc/s) 
			dydt_rec[step+1, reac_count+1] -= dydt_all 
		
	return(dydt_vst) 



'''module is a collection of functions that are used to format chemical schemes, most of them (eg, MCM) are designed for fortran, which have a different format of function names of scientific notations. They must be converted to the format that python can understand'''

import re
import collections

# Remove the comments from the total eqn list and store them in a separate list. Effective 
# comments should start with '//'. 
# This function returns the list of the equations without the 'space' and '\n' markers, along with
# source and sink spcification. In another word, a clean list
def remove_comments(total_list):
    # remove all empty lines and all the '\n' markers at the end of each lines
    naked_list = [line.strip() for line in total_list if line != '\n']
    
    # remove all the comments and use a separate list to store them
    comments = [] # list to store the comments
    comment_marker = r"//.*"
    line_num = 0
    for line_num in range(len(naked_list)):
        if (re.match(comment_marker, naked_list[line_num]) != None):
            comments.append(naked_list[line_num] + '\t' + '(Ln:' + str(line_num) + ')' + '\n')
            naked_list[line_num] = ''
    naked_list = [line for line in naked_list if line != '']

    return naked_list

# convert scientific notation from D-type to E-type
def SN_conversion(RateExp):
	SN_d_regex = r"\d+(D|d)(-?|\+?)\d+"
	RegexList = re.findall(SN_d_regex, RateExp)
	if (RegexList != []): # in case more than one 'D' type instances are in the rate expression
		for i in range(len(RegexList)): # loop the checking until all 'D/d' instances are corrected
			m = re.search(SN_d_regex, RateExp)
			TempSlice = m.group(0) #  get the slice that RE engine has extracted from the rate coef. expression
			SliceStart = m.start() # the position and length are used to replace a part of the expression
			SliceEnd = m.end()
			# replace the 'D/d' with 'e' in the SLICE
			TempSlice = TempSlice.replace('d', 'e')
			TempSlice = TempSlice.replace('D', 'e')
			# reconstruct the rate conef string
			RateExp = RateExp[:SliceStart] + TempSlice + RateExp[SliceEnd:]
			
			# convert '@' marker to '**' (power) marker
			RateExp = RateExp.replace('@', '**')
	if re.search(r'@', RateExp)!=None: # ensure using @ to raise to a power converted
		RateExp = RateExp.replace('@', '**')
	return RateExp

# This function ensures that the function names can be recognized by python
# future development.  The output rate_coef has length equal to the number of
# equations and values giving the reaction rate coefficient.
def convert_rate_mcm(rate_coef):

    # create a list that contains commonly used math functions
    # ('illegal_func_name', func_regex,'legal_func_name')
	math_func_list = [
		('exp', 'numpy.exp'),
        ('EXP', 'numpy.exp'),
        ('dsqrt', 'numpy.sqrt'),
        ('dlog', 'numpy.log'),
        ('LOG', 'numpy.log'),
        ('dabs', 'numpy.abs'),
        ('LOG10', 'numpy.log10')
    ]

    
	# if no change is made, pass the rate expression to a new variable
	# because this new variable is used in the process of photolysis rate J(n)
	new_rate = rate_coef

	# 1. replace the illegal function names
	for func_step in range(len(math_func_list)):

		if (new_rate.find(math_func_list[func_step][0]) != -1):
			new_rate = new_rate.replace(math_func_list[func_step][0], math_func_list[func_step][1]) 
		else:
			continue

	# see if the rate expression contains photolysis rate; convert 'J(n)'/'J<n>' to 'J[n]'
	photolysis_rate_str = re.findall(r"J\(\d+\)|J\<\d+\>", new_rate)
	if (photolysis_rate_str != []): 
		for j_rate in photolysis_rate_str:
			new_j_rate = j_rate.replace('(', '[')
			new_j_rate = new_j_rate.replace('<', '[')
			new_j_rate = new_j_rate.replace(')', ']')
			new_j_rate = new_j_rate.replace('>', ']')
			new_rate = new_rate.replace(j_rate, new_j_rate)


    # The output rate_coef has length equal to the number of
    # equations and values giving the reaction rate coefficient
	return new_rate



'''module to calculate density of particles (kg/m3)'''

import numpy as np

def part_prop(y, num_comp, num_asb, NA, y_mw, y_dens, n0):

	# inputs: ------------------------------------------------------------------
	
	# y - concentration of components in particle size bins 
	#	(molecules/cc (air)), excluding gas and wall
	# num_asb - number of actual size bins (no wall)
	# n0 - particle concentration per size bin now (# particles/cc (air))
	# ---------------------------------------------------------------------------

	y_asmat = (y.reshape(num_asb, num_comp))
	y_asmat = y_asmat.transpose() # species in rows and size bins in columns
	
	# convert molecules/cc (air) to moles/cc (air) and then multiply by molecular weight  
	# to get g/cc (air) of all species in size bins
	y_mass_array = (y_asmat/NA)*y_mw # mass conc. ind. species
	
	mass_fracs = np.zeros((num_comp, num_asb)) # empty matrix for mass fractions
		
	ish = np.array((n0[:, 0]))>0.0 # size bins with particles
	
	# mass fractions of each species including core
	mass_fracs[:, ish] = y_mass_array[:, ish]/((np.sum(y_mass_array, 0))[ish])
	# total density of particles per size bin (g/cm3)
	tot_rho = np.zeros((num_asb))
	tot_rho[ish] = 1.0E-3/((np.sum(mass_fracs/y_dens, 0))[ish])
	# average molecular weight of particles (g/mol)
	avMW = (mass_fracs*y_mw).sum(0)
	
	return(tot_rho, ish, avMW)



'''module to link to calculation of reaction coefficients'''
# this module sets up the final details for calculating
# reaction rate coefficient, which is done via the called module Rate_coeffs 


import numpy as np
import scipy.constants as si


def rrc_calc(RO2_indices, H2O, TEMP, lightm, y, time, lat, lon, act_flux_path, 
		DayOfYear, PInit, photo_par_file, Jlen, tf):

	# ---------------------------------------------
	# inputs:
	
	# TEMP - temperature (K)
	# lightm - whether lights off (0) or on (1)
	# y - component concentrations (molecules/cc (air))
	# time - time of day (for natural light photolysis)
	# lat - latitude
	# lon - longitude
	# act_flux_path - path to file stating actinic flux
	# DayOfYear - day number of the year (1-365)
	# PInit - chamber pressure (Pa)
	# photo_par_file - name of file with with estimates for photolysis absorption
	# 					cross-sections and quantum yields
	# Jlen - number of photolysis reactions
	# tf - sunlight transmission factor
	# ---------------------------------------------
	
	# calculate total RO2 concentration
	if (RO2_indices.size == 0):
		RO2 = 0
	else:
		RO2 = np.sum(y[RO2_indices[:,1]])
		
	# calculate concentrations of third body (M), nitrogen and oxygen
	# calculate gas-phase concentrations of M, N2 and O2 (molecules/cc (air))
	# 1.0e-6 converts from molecules/m3 to molecules/cc
	# R and Avogadro's constant set the same as in atmosphereFunctions.f90 of AtChem2
	M_val = (PInit/(8.3144621*TEMP)*si.N_A)*1.e-6
	# N2 and O2 given the same multiplication as in atmosphereFunctions.f90 of AtChem2
	N2_val = M_val*0.7809
	O2_val = M_val*0.2095
	
	import rate_coeffs
	# calculate the new rate coefficient array (/s) 
	rrc = rate_coeffs.evaluate_rates(RO2, H2O, TEMP, lightm, time, lat, lon, 
						act_flux_path, DayOfYear, M_val, N2_val, 
						O2_val, photo_par_file, Jlen, tf)

		
	return(rrc)




'''module to track particle number size distribution using moving centre size structure (p. 416 of Jacobson 2000)'''

import numpy as np
import scipy.constants as si
import matplotlib.pyplot as plt
from v_check import Vchange_check as Vchange_check

def mov_cen_main(n0, s0, sbn, nc, MW, x, Vol0, t, tmax, C0, MV, 
				Psat, ic_red, res, solv_time, wall_on):


	# input:---------------------------------------------------------
	
	# n0 - particle number concentration per size bin before integration
	# (# particle/cc (air))
	# s0 - volume bounds per size bin (um3) (1st dim.)
	# before a time step over which gas phase reaction and 
	# gas-particle partitioning have occurred (molecules/cc (air))
	# sbn - number of size bins
	# nc - number of components
	# MW - molar weight of components (g/mol)
	# x - original particle size bin radii (um)
	# Vol0 - original volume of size bins (um3) (excluding wall, i.e. volume at mid-point
	# 		between size bin bounds)
	# t - total integration time (s)
	# tmax - maximum integration time (s)
	# tinc_count - count on number of time steps since time interval last required 
	# decreasing
	# C0 - original concentrations (molecules/cc (air))
	# MV - molar volume (cc/mol)
	# Psat - saturation vapour pressures (molecules/cm3 (air))
	# ic_red - flag for time step reduction due to changing initial conditions
	# res - results from every adaptive time step of ode solver, with time steps in rows
	# 		and estimated concentrations of components in all phases in columns 
	#		(molecules/cc (air)) 
	# solv_time - times at which integration solved (s)
	# wall_on - flag for whether wall turned on
	# ---------------------------------------------------------------
	
	NA = si.Avogadro # Avogadro's number (molecules/mol)

	sbn -= wall_on # exclude wall from size bin number
	
	# get new volumes of single particles per size bin and
	# check whether volume change is acceptable
	(redt, t, ic_red, Vnew, tsi) = Vchange_check(res, MV, s0, sbn, NA, 
					n0, nc, solv_time, t, ic_red, Vol0, Psat)
	
	if redt == 1: # repeat integration with new smaller time step
		return(n0, Vol0, C0, x, redt, t, ic_red)
	
	y = np.zeros((nc*(sbn+1+wall_on))) # empty array for holding new concentrations
	# gas and wall concentrations (molecules/cc (air))
	y[0:nc] = res[0:nc]
	if (wall_on == 1):
		y[-nc::] = res[-nc::]
		
	# empty array for holding new particle number concentrations
	N_perbin = np.zeros((sbn, 1))
	
	# if volume condition met, then redistribute particles and components based on
	# the new volume
	for sbi in range(sbn): # size bin loop
		
		sbi_new = sum(Vnew[sbi]>s0[1::]) # index of size bin these particles fit now
		
		# add number concentration (# particles/cc (air))
		N_perbin[sbi_new, 0] += n0[sbi]
		# add components (molecules/cc (air))
		y[((sbi_new+1)*nc):((sbi_new+2)*nc)] += res[((sbi+1)*nc):((sbi+2)*nc)]
	
	# reshape particle-phase concentrations into components in rows and size bins in
	# columns
	num_molec_new = (y[nc:nc*(sbn+1)].reshape(sbn, nc))
	# need to find new volumes of single particles (um3)
	# total volume of components 
	# ((um3 (all particles)/cc (air))/(particle number/cc (air))) 
	# calculation is:
	# divide number of molecules/cc (air) by Na to get moles/cc(air), then 
	# multiply by um3/mol (MV*1.e12) to get ug3 (of each component)/cc (air),
	# then sum volume of components per size bin to get ug3 (all particles)/cc (air)
	ish = N_perbin[:, 0]> 0.0
	Vsing = np.zeros((sbn))
	
	for sbi in range(sbn): # size bin loop
		if (N_perbin[sbi, 0] > 0.): # single particle volume (um3) if particles present
			Vsing[sbi] = np.sum(((y[nc*(sbi+1):nc*(sbi+2)]/(NA*N_perbin[sbi, :]))*(MV[:, 0]*1.e12)), 0)
	
	Vsing[N_perbin[:, 0]<1.e-20] = Vol0[N_perbin[:, 0]<1.e-20] # assume volume at size bin centre
	
	rad = ((3.*Vsing)/(4.*np.pi))**(1./3.) # new radius per size bin (um)
		   
	return(N_perbin, Vsing, y, rad, redt, t, ic_red)



'''module to record required results'''
# keeps track of model variables

import numpy as np
import scipy.constants as si

def rec(save_cnt, trec, yrec, dydt_vst, Cfactor_vst, y, sumt,
	rindx, rstoi, rrc, pindx, pstoi, nprod, 
	nreac, num_sb, num_comp, pconc, core_diss, Psat, kelv_fac, 
	kimt, kw, Cw, act_coeff, Cfactor, Nres_dry, Nres_wet, x2, x,
	MV, H2Oi, Vbou, rbou, wall_on, rbou_rec, seedi):
	
	# inputs: ------------------------------------------------------------
	# save_cnt - count on saving steps
	# trec - the time through simulation record (s)
	# yrec - concentration record (molecules/cc (air))
	# dydt_vst - record of change tendencies (molecules/cc/s)
	# Cfactor_vst - record of the conversion factor for moleculec/ss 
	# to ppb
	# y - concentrations (molecules/cc (air))
	# sumt - cumulative time through simulation (s)
	# rindx - indices of reactants
	# rstoi - stoichiometries of reactants
	# rrc - reaction rate coefficients
	# pindx - indices of products
	# pstoi - stoichiometries of products
	# nprod - number of products
	# nreac - number of reactants
	# num_sb - number of size bins
	# num_comp - number of components
	# pconc - particle concentrations now (#/cc (air))
	# core_diss - dissociation constant of seed
	# Psat - pure component saturation vapour pressure 
	#	(molecules/cm3 (air))
	# kelv_fac - Kelvin factor
	# kimt - partitioning coefficient (/s)
	# kw - gas-wall mass transfer coefficient (/s)
	# Cw - effective absorbing mass of wall (molecules/cc (air))
	# act_coeff - activity coefficients
	# Cfactor - conversion factor from molecules/cc to ppb
	# Nres_dry - number size distributions excluding water (#/cc (air))
	# Nres_wet - number size distributions including water (#/cc (air))
	# x2 - radii of single particle (um)
	# x - radii of single particles per size bin now (um)
	# MV - molar volume of components (um3/mol)
	# H2Oi - index for water
	# Vbou - volume boundaries per size bin (um3)
	# rbou - size bin radius boundaries (um)
	# wall_on - marker for whether wall turned on
	# rbou_rec - size bin radius boundary record (um)
	# seedi - index of seed component
	# --------------------------------------------------------------------

	trec[save_cnt] = sumt # track recording times (s)
	yrec[save_cnt, :] = y # track concentrations (molecules/cc/s)
	# track conversion factor (ppb/molecules/cc)
	Cfactor_vst[save_cnt, 0] = Cfactor
	
	# single particle radius (um) at size bin centre 
	# including contribution of water
	if (num_sb-wall_on > 0):
		x2[save_cnt, :] = x
		# single particle radius boundaries (um) including contribution of water
		rbou_rec[save_cnt, :] = rbou 
	
	# estimate particle number size distributions ----------------------------------
	Vnew = np.zeros((num_sb-wall_on))
	ish = pconc[:, 0]>1.0e-10
	
	if ish.sum()>0:
		# rearrange particle concentrations into size bins in rows, components in columns
		Cn = y[num_comp:num_comp*(num_sb-wall_on+1)].reshape(num_sb-wall_on, num_comp)
		# new volume of single particle per size bin (um3) excluding volume of water	
		Vnew[ish] = (np.sum((Cn[ish, :]/(si.N_A*pconc[ish]))*MV[:, 0]*1.e12, 1)-
				((Cn[ish, H2Oi]/(si.N_A*pconc[ish, 0]))*MV[H2Oi]*1.e12))
		# loop through size bins to find number of particles in each 
		# (# particle/cc (air))
		for Ni in range(0, (num_sb-wall_on)):
			ish = (Vnew>=Vbou[Ni])*(Vnew<Vbou[Ni+1])
			Nres_dry[save_cnt, Ni] = pconc[ish, 0].sum()
		Nres_wet[save_cnt, :] = pconc[:, 0] # record with water
	
	# end of number size distribution part ----------------------------------------


	if len(dydt_vst)>0:
		# record any change tendencies of specified components
		import dydt_rec
		dydt_vst = dydt_rec.dydt_rec(y, rindx, rstoi, rrc, pindx, pstoi, nprod, save_cnt, 
					dydt_vst, nreac, num_sb, num_comp, pconc, core_diss, Psat, kelv_fac, 
					kimt, kw, Cw, act_coeff, seedi)
	save_cnt += 1 # track number of recordings 

	return(trec, yrec, dydt_vst, Cfactor_vst, save_cnt, Nres_dry, Nres_wet, x2, rbou_rec)



'''module to prepare the recording matrices'''
# recording matrices are prepared and initial conditions stored

import numpy as np
import partit_var
import rrc_calc
import cham_up
import scipy.constants as si

# define function
def rec_prep(nrec_step, 
	y, rindx, 
	rstoi, pindx, pstoi, nprod, dydt_vst, nreac, 
	num_sb, num_comp, N_perbin, core_diss, Psat, mfp,
	accom_coeff, y_mw, surfT, R_gas, temp, tempt, NA, 
	y_dens, x, therm_sp, H2Oi, act_coeff, 
	RO2_indx, sumt, Pnow, light_stat, light_time, 
	light_time_cnt, daytime, lat, lon, af_path, 
	dayOfYear, photo_path, Jlen, Cw, kw, Cfactor, tf, 
	light_ad, wall_on, Vbou, tnew, nuc_ad, nucv1, nucv2, nucv3, 
	np_sum, update_stp, update_count, injectt, gasinj_cnt, 
	inj_indx, Ct, pmode, pconc, pconct, seedt_cnt, mean_rad, corei, 
	seed_name, seedVr, lowsize, uppsize, rad0, radn, std, rbou, 
	const_infl_t, infx_cnt, con_infl_C, MV, partit_cutoff, diff_vol, DStar_org, seedi):
	
	# inputs: --------------------------------------------------------
	# nrec_step - number of steps to record on
	# y - initial concentrations (molecules/cc (air))
	# rindx - indices of reactants
	# rstoi - stoichiometries of reactants
	# pindx - indices of products
	# pstoi - stoichiometries of products
	# nprod - number of products
	# dydt_vst - recording for change tendencies
	# nreac - number of reactants
	# num_sb - number of size bins
	# num_comp - number of components
	# N_perbin - particle concentration per size bin (#/cc (air))
	# core_diss - dissociation constant of seed
	# Psat - pure component saturation vapour pressure 
	#	(molecules/cm3 (air))
	# mfp - mean free path (m)
	# accom_coeff - accommodation coefficient
	# y_mw - molecular weight (g/mol)
	# surfT - surface tension (g/s2)
	# R_gas - ideal gas constant (kg.m2.s-2.K-1.mol-1)
	# temp - temperature (K)
	# tempt - times temperatures achieved (s)
	# NA - Avogadro constants (molecules/mol)
	# y_dens - component densities (g/cm3)
	# x - particle radii (um)
	# therm_sp - thermal speed (m/s)
	# H2Oi - index for water
	# act_coeff - activity coefficient
	# RO2_indx - index of peroxy radicals
	# sumt - cumulative time through simulation (s)
	# Pnow - chamber pressure (Pa)
	# light_stat - light status
	# light_time - times corresponding to light status
	# light_time_cnt - count on light status array
	# daytime - start time of simulation (s)
	# lat - latitude
	# lon - longitude
	# af_path - path to actinic flux file
	# dayOfYear - day of year
	# photo_path - path to photochemistry file
	# Jlen - length of photochemistry equations
	# Cw - effective absorbing mass of wall (molecules/cc (air))
	# kw - gas-wall partitioning coefficient (/s)
	# Cfactor - conversion factor (ppb/molecules/cc)
	# tf - transmission factor for natural sunlight
	# light_ad - marker for whether to adapt time interval to 
	#		changing natural light intensity
	# wall_on - marker for whether wall present
	# Vbou - volume boundary of particle size bins (um3)
	# tnew - the proposed integration time interval (s)
	# nuc_ad - marker for whether to adapt time step based on 
	#	nucleation
	# nucv1 - nucleation parameter one
	# nucv2 - nucleation parameter two
	# nucv3 - nucleation parameter three
	# np_sum - cumulative number concentration of new 
	#	particles so far (#/cc (air))
	# update_stp - time interval between operator-split processes
	# update_count - count on time since last operator-split (s)
	# injectt - time of instantaneous injection of components (s)
	# gasinj_cnt - count on injection times of components
	# inj_indx - index of component(s) being injected after 
	#	experiment start
	# Ct - concentration(s) (ppb) of component(s) injected 
	# instantaneously after experiment start
	# pmode - whether number size distributions expressed as modes or explicitly
	# pconc - concentration of injected particles (#/cc (air))
	# pconct - times of particle injection (s)
	# seedt_cnt - count on injection of seed particles
	# mean_rad - mean radius for particle number size 
	#	distribution (um)
	# corei - index of core component
	# seedVr - volume ration of component(s) comprising seed particles
	# seed_name - name(s) of component(s) comprising seed particles
	# seedVr - volume ratio of component(s) comprising seed particles
	# lowsize - lower size bin boundary (um)
	# uppsize - upper size bin boundary (um)
	# rad0 - initial radius at size bin centres (um)
	# radn - current radius at size bin centres (um)	
	# std - standard deviation for injected particle number size 
	#	distributions
	# rbou - size bin radius bounds (um)
	# const_infl_t - times for constant influxes (s)
	# infx_cnt - count on constant influx occurrences
	# con_infl_C - influx rates of components with constant influx 
	#	(ppb/s)
	# MV - molar volume (cc/mol)
	# partit_cutoff - the product of saturation vapour pressure and
	#		activity coefficient above which gas-particle
	#		partitioning assumed zero (Pa)
	# diff_vol - diffusion volume of components according to Fuller et al. (1969)
	# DStar_org - gas-phase diffusion coefficient of components (cm2/s)
	# seedi - index of seed component(s)
	# ----------------------------------------------------------------

	# array to record time through simulation (s)
	trec = np.zeros((nrec_step))
	# array to record concentrations with time (molecules/cc/s)
	yrec = np.zeros((nrec_step, len(y)))
	yrec[0, :] = y # record initial concentrations (molecules/cc/s)
	# array to record conversion factor
	Cfactor_vst = np.zeros((nrec_step, 1))
	Cfactor_vst[0] = Cfactor

	# arrays to record particle radii and number size distributions	
	if ((num_sb-wall_on) > 0):
		x2 = np.zeros((nrec_step, (num_sb-wall_on)))
		Nres_dry = np.zeros((nrec_step, (num_sb-wall_on)))
		Nres_wet = np.zeros((nrec_step, (num_sb-wall_on)))
		rbou_rec = np.zeros((nrec_step, (num_sb-wall_on+1)))
	else:
		x2 = 0.
		Nres_dry = 0.
		Nres_wet = 0.
		rbou_rec = 0.

	# update chamber variables
	[temp_now, Pnow, lightm, light_time_cnt, tnew, ic_red, update_stp, 
		update_count, Cinfl_now, seedt_cnt, Cfactor, infx_cnt, 
		gasinj_cnt, DStar_org] = cham_up.cham_up(sumt, temp, tempt, 
		Pnow, light_stat, light_time, light_time_cnt, light_ad, 0, 
		nuc_ad, nucv1, nucv2, nucv3, np_sum, 
		update_stp, update_count, lat, lon, dayOfYear, photo_path, 
		af_path, injectt, gasinj_cnt, inj_indx, Ct, pmode, pconc, pconct, 
		seedt_cnt, num_comp, y, N_perbin, mean_rad, corei, seedVr, seed_name, 
		lowsize, uppsize, num_sb, MV, rad0, radn, std, y_dens, H2Oi, rbou, 
		const_infl_t, infx_cnt, con_infl_C, wall_on, Cfactor, seedi, diff_vol, DStar_org)
	
	
	if (num_sb-wall_on)>0: # if particles present
		
		# update partitioning variables
		[kimt, kelv_fac] = partit_var.kimt_calc(y, mfp, num_sb, num_comp, accom_coeff, y_mw,   
		surfT, R_gas, temp_now, NA, y_dens*1.e3, N_perbin, 
		x.reshape(1, -1)*1.0e-6, Psat, therm_sp, H2Oi, act_coeff, wall_on, 1, partit_cutoff, 
		Pnow, DStar_org)
		
		# single particle radius (um) at size bin centre 
		# including contribution of water
		x2[0, :] = x

		# single particle size bin bounds including water (um)
		rbou_rec[0, :] = rbou
		
		# estimate particle number size distributions ----------------------------------
		Vnew = np.zeros((num_sb-wall_on))
		ish = N_perbin[:, 0]>1.0e-10
		
		if ish.sum()>0:
			# rearrange particle concentrations into size bins in rows, components in columns
			Cn = y[num_comp:num_comp*(num_sb-wall_on+1)].reshape(num_sb-wall_on, num_comp)
			# new volume of single particle per size bin (um3) excluding volume of water	
			Vnew[ish] = (np.sum((Cn[ish, :]/(si.N_A*N_perbin[ish]))*MV[:, 0]*1.e12, 1)-
					((Cn[ish, H2Oi]/(si.N_A*N_perbin[ish, 0]))*MV[H2Oi, 0]*1.e12))
			# loop through size bins to find number of particles in each 
			# (# particle/cc (air))
			for Ni in range(0, (num_sb-wall_on)):
				ish = (Vnew>=Vbou[Ni])*(Vnew<Vbou[Ni+1])
				Nres_dry[0, Ni] = N_perbin[ish, 0].sum()
			Nres_wet[0, :] = N_perbin[:, 0] # record with water
		
		# end of number size distribution part ----------------------------------------

	else: # fillers
		kimt = kelv_fac = 0.
	
	# update reaction rate coefficients
	rrc = rrc_calc.rrc_calc(RO2_indx, 
			y[H2Oi], temp_now, lightm, y, daytime+sumt, 
			lat, lon, af_path, dayOfYear, Pnow, 
			photo_path, Jlen, tf)

	if len(dydt_vst)>0:		
		# record any change tendencies of specified components
		import dydt_rec
		dydt_vst = dydt_rec.dydt_rec(y, rindx, rstoi, rrc, pindx, pstoi, nprod, 0, 
					dydt_vst, nreac, num_sb, num_comp, N_perbin, core_diss, Psat, kelv_fac, 
					kimt, kw, Cw, act_coeff, corei)


	return(trec, yrec, dydt_vst, Cfactor_vst, Nres_dry, Nres_wet, x2, seedt_cnt, rbou_rec, Cfactor, infx_cnt)



'''module for calculating photolysis rates'''
# can use ambient sunlight expression or artificial chamber lights for calculating a
# photolysis rate that is fixed when the lights are on

import scipy
import os
import numpy as np
import requests, zipfile, io # for downloading
import shutil
from lamp_photo import lamp_photo
import zenith

def PhotolysisCalculation(time, lat, lon, TEMP, act_flux_path, DayOfYear, photo_par_file,
							Jlen, tf):

	# inputs:-----------------------------------------------------------------------------	
	# act_flux_path - name of path to file containing known actinic flux (only used if 
	#				lights on inside chamber)
	# DayOfYear - number of days through the calendar year
	# photo_par_file - name of file containing estimates for wavelength-dependent
	# 					absorption cross-sections and quantum yields
	# Jlen - number of photolysis reactions
	# tf - the transmission factor (for natural light intensity)
	# ------------------------------------------------------------------------------------
	
	(secx, cosx) = zenith.zenith(time, lat, lon, DayOfYear)
	
	J = np.zeros((Jlen)) # prepare output
    
	# if using MCM and natural light
	cwd = os.getcwd() # address of current working directory

	if (photo_par_file == str(cwd + '/PyCHAM/photofiles/MCMv3.2') and act_flux_path == 'no'):
		#J          L           M          N
		J[1]=6.073E-05*cosx**(1.743)*np.exp(-1.0*0.474*secx)
		J[2]=4.775E-04*cosx**(0.298)*np.exp(-1.0*0.080*secx)
		J[3]=1.041E-05*cosx**(0.723)*np.exp(-1.0*0.279*secx)
		J[4]=1.165E-02*cosx**(0.244)*np.exp(-1.0*0.267*secx)
		J[5]=2.485E-02*cosx**(0.168)*np.exp(-1.0*0.108*secx)
		J[6]=1.747E-01*cosx**(0.155)*np.exp(-1.0*0.125*secx)
		J[7]=2.644E-03*cosx**(0.261)*np.exp(-1.0*0.288*secx)
		J[8]=9.312E-07*cosx**(1.230)*np.exp(-1.0*0.307*secx)
		J[11]=4.642E-05*cosx**(0.762)*np.exp(-1.0*0.353*secx)
		J[12]=6.853E-05*cosx**(0.477)*np.exp(-1.0*0.323*secx)
		J[13]=7.344E-06*cosx**(1.202)*np.exp(-1.0*0.417*secx)
		J[14]=2.879E-05*cosx**(1.067)*np.exp(-1.0*0.358*secx)
		J[15]=2.792E-05*cosx**(0.805)*np.exp(-1.0*0.338*secx)
		J[16]=1.675E-05*cosx**(0.805)*np.exp(-1.0*0.338*secx)
		J[17]=7.914E-05*cosx**(0.764)*np.exp(-1.0*0.364*secx)
		J[18]=1.140E-05*cosx**(0.396)*np.exp(-1.0*0.298*secx)
		J[19]=1.140E-05*cosx**(0.396)*np.exp(-1.0*0.298*secx)
		J[21]=7.992E-07*cosx**(1.578)*np.exp(-1.0*0.271*secx)
		J[22]=5.804E-06*cosx**(1.092)*np.exp(-1.0*0.377*secx)
		J[23]=1.836E-05*cosx**(0.395)*np.exp(-1.0*0.296*secx)
		J[24]=1.836E-05*cosx**(0.395)*np.exp(-1.0*0.296*secx)
		J[31]=6.845E-05*cosx**(0.130)*np.exp(-1.0*0.201*secx)
		J[32]=1.032E-05*cosx**(0.130)*np.exp(-1.0*0.201*secx)
		J[33]=3.802E-05*cosx**(0.644)*np.exp(-1.0*0.312*secx)
		J[34]=1.537E-04*cosx**(0.170)*np.exp(-1.0*0.208*secx)
		J[35]=3.326E-04*cosx**(0.148)*np.exp(-1.0*0.215*secx)
		J[41]=7.649E-06*cosx**(0.682)*np.exp(-1.0*0.279*secx)
		J[51]=1.588E-06*cosx**(1.154)*np.exp(-1.0*0.318*secx)
		J[52]=1.907E-06*cosx**(1.244)*np.exp(-1.0*0.335*secx)
		J[53]=2.485E-06*cosx**(1.196)*np.exp(-1.0*0.328*secx)
		J[54]=4.095E-06*cosx**(1.111)*np.exp(-1.0*0.316*secx)
		J[55]=1.135E-05*cosx**(0.974)*np.exp(-1.0*0.309*secx)
		J[56]=7.549E-06*cosx**(1.015)*np.exp(-1.0*0.324*secx)
		J[57]=3.363E-06*cosx**(1.296)*np.exp(-1.0*0.322*secx)
		J[61]=7.537E-04*cosx**(0.499)*np.exp(-1.0*0.266*secx)

		J = J*tf

	# from MAC spectral analysis and Mainz database (xsproc.py)
# 	J[1] = 2.3706768705670786e-05
# 	J[2] = 7.128976494635883e-05
# 	J[3] = 1.26339378672e-05
# 	J[4] = 0.0011216654096221574
# 	J[5] = 0.0024080663555999995
# 	J[6] = 0.02122656504799999
# 	J[7] = 0.0002830012299999995
# 	J[8] = 2.8579387129999973e-07
# 	J[11] = 1.3984589959570004e-05
# 	J[12] = 6.155370761783531e-06
# 	J[15] = 3.657979548045003e-05
# 	J[21] = 2.4121216268999992e-06
# 	J[22] = 1.0067723552799998e-05
# 	J[31] = 8.998612475770001e-07
# 	J[32] = 3.710891288701262e-06
# 	J[33] = 8.468352055021745e-06
# 	J[34] = 1.039182783128049e-05
# 	J[35] = 3.1229494794723276e-05
# 	J[41] = 0.014555553853016991
	
	# if a file path for the chamber's actinic flux is supplied, then use the MCM 
	# equations for photolysis rates as a function of wavelength-dependent actinic flux
	# this is done if artificial lights are turned on
	if (act_flux_path != 'no'):
		# using MCM recommended values
# 		folder_url = "http://mcm.leeds.ac.uk/MCMv3.3.1/parameters/photolysis/MCMv3.2photolysis.zip"
# 		r = requests.get(folder_url) # create HTTP response object
# 		if r.ok==0: # check object is ready
# 			print('Photolysis information not found at the folowing URL, please check and update in PhotolysisRates.py: http://mcm.leeds.ac.uk/MCMv3.3.1/parameters/photolysis/MCMv3.2photolysis.zip')
# 		z = zipfile.ZipFile(io.BytesIO(r.content)) #  download object
# 		z.extractall('PyCHAM/MCMphotofiles') # creates folder and stores files there
		
		# call on MCM_photo module to process photolysis files and estimate J values
		J = lamp_photo(photo_par_file, J, TEMP, act_flux_path)
	
		# remove photolysis information folder
# 		cwd = os.getcwd() # address of current working directory
# 		if os.path.isdir(cwd + '/PyCHAM/MCMphotofiles/' + Photo_name[0]): # check if there is an existing Photo_name folder
# 			def handleRemoveReadonly(func, path, exc):
# 				excvalue = exc[1]
# 				if not os.access(path, os.W_OK):
# 					# Is the error an access error ?
# 					os.chmod(path, stat.S_IWUSR)
# 					func(path)
# 				else:
# 					raise
# 			shutil.rmtree(cwd + '/PyCHAM/MCMphotofiles/' + Photo_name[0], ignore_errors=False, onerror=handleRemoveReadonly) # remove existing folder, onerror will change permission of directory if needed

	
	return(J)



'''preparing the matrices for aqueous-phase reactions'''
# mostly just adjusting indices to account for number of size bins

import numpy as np

def aq_mat_prep(rindx, rstoi, pindx, pstoi, reac_coef, 
	nprod, jac_stoi, njac,
	jac_den_indx, jac_indx, 				
	y_arr, y_rind, uni_y_rind, y_pind, 
	uni_y_pind, reac_col, prod_col, rstoi_flat, pstoi_flat, 
	rr_arr, rr_arr_p, num_sb, wall_on, num_eqn, comp_num):

	# inputs: --------------------------------------------------------------
	# rindx - index for reactants
	# rstoi - stoichiometries of reactants
	# pindx - index for products
	# pstoi - stoichiometries of products
	# nprod - number of products per equation
	# njac - number of Jacobian elements affected per equation
	# jac_stoi - stoichiometries for jacobian
	# jac_den_indx_aq - indices for Jacobian denominators per equation
	# y_arr - indices of array for holding reactant concentrations
	# y_rind - indices of concentration array for reactants
	# uni_y_rind - indices of reactants in concentration array
	# y_pind - indices of concentration array for products
	# uni_y_pind - indices of products in concentration array
	# reac_col - columns for sparse matrix 
	# prod_col - columns for sparse matrix
	# rstoi_flat - flattened reactant stoichiometries
	# pstoi_flat - falttened product stoichiometries
	# rr_arr - reaction rate array indices for reactants
	# rr_arr_p - reaction rate indices for products
	# num_sb - number of size bins
	# wall_on - marker for whether wall on or off
	# num_eqn - number of aqueous reactions
	# comp_num - number of components
	# ----------------------------------------------------------------------

	num_asb = (num_sb-wall_on) # number of particle size bins
	# shapes of arrays
	rindxs = rindx.shape
	pindxs = pindx.shape
	y_arrl = len(y_arr)
	y_rindl = len(y_rind)
	y_pindl = len(y_pind)
	rr_arrl = len(rr_arr)
	rr_arr_pl = len(rr_arr_p)
	reac_coll = len(reac_col)
	prod_coll = len(prod_col)
	uni_y_rindl = len(uni_y_rind)
	uni_y_pindl = len(uni_y_pind)
	
	# tile where possible (ahead of vectorised operation)
	rstoi = np.tile(rstoi, (num_asb, 1))
	pstoi = np.tile(pstoi, (num_asb, 1))
	rstoi_flat = np.tile(rstoi_flat, (1, num_asb))
	pstoi_flat = np.tile(pstoi_flat, (1, num_asb))
	
	for sbi in range(num_sb-wall_on): # size bin loop
		
		if (sbi == 0): # first size bin - account for gas-phase indices prior
			rindx += (sbi+1)*(comp_num+2)
			pindx += (sbi+1)*(comp_num+2)
			y_rind = y_rind+(sbi+1)*(comp_num+2)
			y_pind = y_pind+(sbi+1)*(comp_num+2)
			uni_y_rind = uni_y_rind+(comp_num+2)
			uni_y_pind = uni_y_pind+(comp_num+2)

		if (sbi > 0): # larger size bin
			rindx = np.append(rindx, rindx[0:rindxs[0], 0:rindxs[1]]+(sbi)*(comp_num+2), axis=0)
			pindx = np.append(pindx, pindx[0:pindxs[0], 0:pindxs[1]]+(sbi)*(comp_num+2), axis=0)
			y_arr = np.append(y_arr, y_arr[0:y_arrl]+(sbi)*(max(y_arr[0:y_arrl])+1), axis=0)
			y_rind = np.append(y_rind, y_rind[0:y_rindl]+(sbi)*(comp_num+2))
			y_pind = np.append(y_pind, y_pind[0:y_pindl]+(sbi)*(comp_num+2))
			rr_arr = np.append(rr_arr, rr_arr[0:rr_arrl]+(sbi)*(max(rr_arr[0:rr_arrl])+1), axis=0)
			rr_arr_p = np.append(rr_arr_p, rr_arr_p[0:rr_arr_pl]+(sbi)*(max(rr_arr_p[0:rr_arr_pl])+1), axis=0)
			reac_col = np.append(reac_col, reac_col[reac_coll-1]+reac_col[-reac_coll+1::])
			prod_col = np.append(prod_col, prod_col[prod_coll-1]+prod_col[-prod_coll+1::])
			uni_y_rind = np.append(uni_y_rind, uni_y_rind[0:uni_y_rindl]+(sbi)*(comp_num+2), axis=0)
			uni_y_pind = np.append(uni_y_pind, uni_y_pind[0:uni_y_pindl]+(sbi)*(comp_num+2), axis=0)
			jac_stoi = np.append(jac_stoi, np.zeros((jac_stoi.shape[0], int(max(njac)))), axis=1)
			jac_den_indx = np.append(jac_den_indx, np.zeros((jac_den_indx.shape[0], int(max(njac)))), axis=1)
			for eqi in range(njac.shape[0]): # equation loop
				jac_stoi[eqi, int(sbi*njac[eqi]):int((sbi+1)*njac[eqi])] = jac_stoi[eqi, 0:int(njac[eqi])]
				jac_den_indx[eqi, int(sbi*njac[eqi]):int((sbi+1)*njac[eqi])] = jac_den_indx[eqi, 0:int(njac[eqi])]+(sbi)*(comp_num+2)
	
	# account for size bins
	njac = njac*(num_sb-wall_on)
	
	# ensure integer type
	njac = njac.astype(int)
	jac_den_indx = jac_den_indx.astype(int)

	return(rindx, rstoi, pindx, pstoi, reac_coef, 
		nprod, jac_stoi, njac,
		jac_den_indx, jac_indx, 				
		y_arr, y_rind, uni_y_rind, y_pind, 
		uni_y_pind, reac_col, prod_col, rstoi_flat, pstoi_flat, 
		rr_arr, rr_arr_p)



'''module to estimate component volatilities and liquid densities'''
# This module is responsible for
# setting key properties of components, including liquid-phase saturation vapour pressures
# and liquid-phase densities.  It does this using either UManSysProp (default), or with
# user settings

import numpy as np
import sys
import os
from git import Repo
import shutil
import scipy.constants as si
import errno
import stat

def prop_calc(comp_list, Pybel_objects, TEMP, H2Oi, num_comp, Psat_water, vol_Comp, 
				volP, testf, corei, pconc, umansysprop_update, core_dens, spec_namelist,
				ode_gen_flag, nuci, nuc_comp, num_asb, dens_comp, dens, seed_name):

	# inputs: ------------------------------------------------------------
	# comp_list - array of SMILE strings for components 
	# (omitting water and core, if present)
	# Pybel_objects - list of Pybel objects representing the species in comp_list
	# (omitting water and core, if present)
	# TEMP - temperature (K) in chamber at time function called
	# vol_Comp - names of components (corresponding to those in chemical scheme file)
	# 			that have vapour pressures manually set in volP
	# testf - flag for whether in normal mode (0) or testing mode (1)
	# corei - index of seed particle component
	# pconc - initial number concentration of particles (#/cc (air))
	# umansysprop_update - marker for cloning UManSysProp so that latest version used
	# core_dens - density of core material (g/cc (liquid/solid density))
	# spec_namelist - list of components' names in chemical equation file
	# ode_gen_flag - whether or not called from middle or ode_gen
	# nuci - index of nucleating component
	# nuc_comp - name of nucleating component
	# num_asb - number of actual size bins (excluding wall)
	# dens_comp - chemical scheme names of components with manually assigned 
	# 	densities
	# dens - manually assigned densities (g/cc)
	# seed_name - chemical scheme name(s) of component(s) comprising seed particles
	# ------------------------------------------------------------
	
	
	
	if testf==1:
		return(0,0,0) # return dummies
		
	cwd = os.getcwd() # address of current working directory
	if umansysprop_update == 1:
		print('Cloning latest version of UManSysProp via prop_calc module')
		# download latest version of umansysprop
		
		# check if there is an existing umansysprop folder
		if os.path.isdir(cwd + '/umansysprop'): 
			def handleRemoveReadonly(func, path, exc):
				excvalue = exc[1]
				if not os.access(path, os.W_OK):
					# Is the error an access error ?
					os.chmod(path, stat.S_IWUSR)
					func(path)
				else:
					raise
			# remove existing folder, onerror will change permission of directory if 
			# needed
			shutil.rmtree(cwd + '/umansysprop', ignore_errors=False, onerror=handleRemoveReadonly)
		
		git_url = 'https://github.com/loftytopping/UManSysProp_public.git'
		Repo.clone_from(git_url, (cwd + '/umansysprop'))
	
	# point to umansysprop folder
	sys.path.insert(1, (cwd + '/umansysprop')) # address for updated version
	
	from umansysprop import boiling_points
	from umansysprop import vapour_pressures
	from umansysprop import liquid_densities

	NA = si.Avogadro # Avogadro's number (molecules/mol)
	y_dens = np.zeros((num_comp, 1)) # components' liquid density (kg/m3)
	# vapour pressures of components, ensures any seed component called 
	# core has zero vapour pressure
	Psat = np.zeros((1, num_comp))

	
	if (ode_gen_flag == 0): # estimate densities if called from middle
		
		for i in range (num_comp): # loop through components
			
			# density estimation ---------------------------------------------------------
			if (i == H2Oi): # liquid-phase density of water
				y_dens[i] = 1.*1.E3 # (kg/m3 (particle))
				continue
			# core component properties
			if (i == corei[0]): # density of core
				y_dens[i] = core_dens*1.e3 # core density (kg/m3 (particle))
				continue
			if comp_list[i] == '[HH]': # omit H2 as unliked by liquid density code
				# liquid density code does not like H2, so manually input kg/m3
				y_dens[i] = 1.0e3
			else:
				# density (convert from g/cc to kg/m3)
				y_dens[i] = liquid_densities.girolami(Pybel_objects[i])*1.0E3
			# ----------------------------------------------------------------------------
		
	# account for any manually assigned component densities (kg/m3)
	if (len(dens_comp) > 0  and ode_gen_flag == 0):
		for i in range (len(dens_comp)):
			# index of component in list of components
			dens_indx = spec_namelist.index(dens_comp[i])
			y_dens[dens_indx] = dens[i]
	
	# estimate vapour pressures (log10(atm))
	for i in range (num_comp):
		
		if (i == corei[0]): # if this component is 'core'
			# core component not included in Pybel_objects, continuing
			# here means its vapour pressure is 0 Pa, which is fine
			continue
		
		# water vapour pressure already given by Psat_water (log10(atm))
		if i == H2Oi:
			Psat[0, i] = Psat_water
			continue # water not included in Pybel_objects
		
		# vapour pressure (log10 atm) (# eq. 6 of Nannoolal et al. (2008), with dB of 
		# that equation given by eq. 7 of same reference)
		Psat[0, i] = ((vapour_pressures.nannoolal(Pybel_objects[i], TEMP, 
						boiling_points.nannoolal(Pybel_objects[i]))))
	
	ish = (Psat == 0.)
	Psat = (np.power(10.0, Psat)*101325.0) # convert to Pa from atm
	# retain low volatility where wanted following unit conversion
	Psat[ish] = 0.
	
	# manually assigned vapour pressures (Pa)
	if (len(vol_Comp) > 0 and ode_gen_flag == 0):
		for i in range (len(vol_Comp)):
			# index of component in list of components
			vol_indx = spec_namelist.index(vol_Comp[i])
			Psat[0, vol_indx] = volP[i]

	# ensure if nucleating component is core that it is involatile
	if (nuc_comp == 'core'):
		Psat[0, nuci] = 0.
	
	Psat_Pa = np.zeros((1, num_comp)) # for storing vapour pressures in Pa (Pa)
	Psat_Pa[0, :] = Psat[0, :]
    
	# convert saturation vapour pressures from Pa to molecules/cc (air) using ideal
	# gas law, R has units cc.Pa/K.mol
	Psat = Psat*(NA/((si.R*1.e6)*TEMP))
	# now, in preparation for ode solver, repeat over number of size bins
	if (num_asb > 0):
		Psat = np.repeat(Psat, num_asb, axis=0)
	
	return(Psat, y_dens, Psat_Pa)



'''implement the fully-moving size structure'''
# following Section 13.5.2 of Jacobson (2005) (Fundamentals of Atmospheric Modeling),
# the full-moving size structure revalues the size of particles per size bin
# following gas-particle partitioning

import scipy.constants as si
import numpy as np

def fullmov(num_sb, n0, num_comp, Cp, MV, Vol0, Vbou, rbou): # define module

	# inputs: -----------------------------------------------------
	# num_sb - number of size bins excluding wall (if present)
	# n0 - particle number concentration per size bin
	# num_comp - number of components
	# Cp - particle-phase component concentrations (molecules/cc (air))
	# MV - molar volume per component (um3/mol)
	# Vol0 - initial volume per size bin at bin centre (um3)
	# Vbou - volume bounds between size bins (um3)
	# rbou - radius bounds (um)
	# -------------------------------------------------------------

	# particle-phase concentrations	(molecules/cc (air))
	Cp = np.transpose(Cp.reshape(num_sb, num_comp))
	NA = si.Avogadro # Avogadro's number (molecules/mol)
	ish = n0[:, 0]>0. # size bins containing particles


	nmolC = np.zeros((num_comp, ish.sum())) # empty array for molar concentration
	# number of moles of each component in a single particle (mol/cc (air))
	nmolC[:, :] = ((Cp[:, ish]/(NA*n0[ish, 0])))

	Vnew = np.zeros((num_sb)) # empty array for new volumes
	MVrep = np.repeat(MV, num_sb, axis=1)
	# new volume of single particle per size bin (um3), including volume of water
	Vnew[ish] = np.sum(nmolC*MVrep[:, ish], axis=0)
	
	# if no particles in a size bin, assign starting volume (um3)
	Vnew[n0[:, 0]<=1.0e-10] = Vol0[0::][n0[:, 0]<=1.0e-10]

	# flatten particle-phase concentrations (molecules/cc (air))
	Cp = np.ravel(np.transpose(Cp))

	Vnew_ord = sorted(Vnew) # arrange size bins in ascending size order
	Cpn = np.zeros((len(Cp)))
	nn = np.zeros((n0.shape[0], n0.shape[1]))
	# arrange particle-phase concentrations and particle number concentrations similarly
	for i in range(len(Vnew)):
		ind = Vnew_ord.index(Vnew[i]) # new index
		Cpn[(ind)*num_comp:(ind+1)*num_comp] = Cp[(i)*num_comp:(i+1)*num_comp]
		nn[ind] = n0[i]
	
	Vnew_ord = np.array((Vnew_ord)) # transform list to numpy array

	# new volume bounds (um3) between size bin	
	Vbou[1:-1] = Vnew_ord[0:-1]+(Vnew_ord[1::]-Vnew_ord[0:-1])/2.

	# new radii per size bin (um)
	x = ((3.*Vnew_ord)/(4.*np.pi))**(1./3.)
		
	# new radius bounds (um)
	rbou = ((3.*Vbou)/(4.*np.pi))**(1./3.)


	return(Vnew_ord, x, Cpn, nn, Vbou, rbou)
